"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.getSign = exports.isHex = exports.normalizeNumber = exports.newlineKindAt = exports.findQuotes = exports.regexMatchLength = exports.isNumeric = exports.graphemePositionFrom = exports.columnNumberFrom = exports.ensureQuoted = exports.assertGetFormatted = exports.NewlineKind = exports.graphemeSplitter = void 0;
// tslint:disable-next-line: no-require-imports
const GraphemeSplitter = require("grapheme-splitter");
const _1 = require(".");
exports.graphemeSplitter = new GraphemeSplitter();
var NewlineKind;
(function (NewlineKind) {
    NewlineKind["SingleCharacter"] = "SingleCharacter";
    NewlineKind["DoubleCharacter"] = "DoubleCharacter";
})(NewlineKind = exports.NewlineKind || (exports.NewlineKind = {}));
// A quick and dirty way to do string formatting.
// Does not handle any escaping.
function assertGetFormatted(template, args) {
    let result = template;
    for (const [key, value] of args.entries()) {
        const formatKey = `{${key}}`;
        _1.Assert.isTrue(template.indexOf(formatKey) !== -1, `unknown formatKey in template`, { template, formatKey });
        result = result.replace(formatKey, value);
    }
    return result;
}
exports.assertGetFormatted = assertGetFormatted;
function ensureQuoted(text) {
    if (findQuotes(text, 0)) {
        return text;
    }
    return `"${text.replace(/"/g, '""')}"`;
}
exports.ensureQuoted = ensureQuoted;
function columnNumberFrom(text, requiredCodeUnit) {
    const graphemes = exports.graphemeSplitter.splitGraphemes(text);
    let columnNumber = 0;
    let summedCodeUnits = 0;
    for (const grapheme of graphemes) {
        if (summedCodeUnits === requiredCodeUnit) {
            return columnNumber;
        }
        else {
            summedCodeUnits += grapheme.length;
            columnNumber += 1;
        }
    }
    const details = {
        text,
        requiredCodeUnit,
    };
    throw new _1.CommonError.InvariantError(`no columnNumber can be generated for required codeUnit`, details);
}
exports.columnNumberFrom = columnNumberFrom;
function graphemePositionFrom(text, lineCodeUnit, lineNumber, codeUnit) {
    return {
        lineCodeUnit,
        lineNumber,
        columnNumber: columnNumberFrom(text, lineCodeUnit),
        codeUnit,
    };
}
exports.graphemePositionFrom = graphemePositionFrom;
function isNumeric(text) {
    return regexMatchLength(_1.Pattern.Numeric, text, 0) === text.length;
}
exports.isNumeric = isNumeric;
function regexMatchLength(pattern, text, index) {
    pattern.lastIndex = index;
    const matches = pattern.exec(text);
    return matches !== null && matches.index === index ? matches[0].length : undefined;
}
exports.regexMatchLength = regexMatchLength;
// Attempt to find quoted text at the given starting index.
// Return undefined if the starting index isn't a quote or if there is no closing quotes.
function findQuotes(text, indexStart) {
    const textLength = text.length;
    if (
    // If it doesn't start with a quote
    text[indexStart] !== '"' ||
        // or indexStart was the last character
        indexStart + 1 === textLength) {
        return undefined;
    }
    let continueMatching = true;
    let index = indexStart + 1;
    while (continueMatching) {
        const chr1 = text[index];
        if (chr1 === '"') {
            const chr2 = text[index + 1];
            if (chr2 !== '"') {
                continueMatching = false;
                index += 1;
                continue;
            }
            else {
                index += 2;
            }
        }
        index += 1;
        if (index >= textLength) {
            continueMatching = false;
        }
    }
    // If no valid end quote was found
    if (index >= textLength && text[textLength - 1] !== `"`) {
        return undefined;
    }
    return {
        indexStart,
        indexEnd: index,
        quoteLength: index - indexStart,
    };
}
exports.findQuotes = findQuotes;
function newlineKindAt(text, index) {
    const chr1 = text[index];
    switch (chr1) {
        case `\u000d`: {
            const chr2 = text[index + 1];
            return chr2 === `\u000a` ? NewlineKind.DoubleCharacter : NewlineKind.SingleCharacter;
        }
        case `\u000a`:
        case `\u0085`:
        case `\u2028`:
            return NewlineKind.SingleCharacter;
        default:
            return undefined;
    }
}
exports.newlineKindAt = newlineKindAt;
function normalizeNumber(text) {
    const [isPositive, charOffset] = getSign(text);
    const allButUnaryOperators = text.slice(charOffset);
    if (isHex(allButUnaryOperators)) {
        return isPositive ? allButUnaryOperators.toLocaleLowerCase() : `-${allButUnaryOperators.toLocaleLowerCase()}`;
    }
    else if (regexMatchLength(_1.Pattern.Numeric, allButUnaryOperators, 0) !== allButUnaryOperators.length) {
        return undefined;
    }
    else {
        return isPositive ? allButUnaryOperators : `-${allButUnaryOperators}`;
    }
}
exports.normalizeNumber = normalizeNumber;
function isHex(text) {
    return regexMatchLength(_1.Pattern.Hex, text, 0) === text.length;
}
exports.isHex = isHex;
function getSign(text) {
    let char = text[0];
    let charOffset = 0;
    let isPositive = true;
    while (char === "-" || char === "+") {
        isPositive = char === "-" ? !isPositive : isPositive;
        charOffset += 1;
        char = text[charOffset];
    }
    return [isPositive, charOffset];
}
exports.getSign = getSign;
//# sourceMappingURL=stringUtils.js.map