"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.NoOpTraceManagerInstance = exports.NoOpTraceManager = exports.BenchmarkTraceManager = exports.ReportTraceManager = exports.TraceManager = exports.NoOpTraceInstance = exports.NoOpTrace = exports.BenchmarkTrace = exports.Trace = exports.TraceConstant = void 0;
// tslint:disable-next-line: no-require-imports
const performanceNow = require("performance-now");
// phase:
//      Static string.
//      Examples: 'Lex', 'Parse', 'StaticAnalysis'
// task:
//      Static string.
//      Something that falls under the given phase. Recommended to be the name of a function.
//      Examples: 'readNumericLiteral' for the phase 'Lex', or 'parseRecord' for the phase 'Parse'.
// id:
//      Dynamic number.
//      Used to help identify uniqueness. Auto incrementing integer.
// correlationId:
//      Can be undefined. If truthy then it refers to a previously generated 'id' number.
//      Used to track execution flow.
// message:
//      Static string.
//      Identifies what portion of a task you're in.
//      Examples: 'entry', 'partialEvaluation', 'traceExit'.
// details:
//      Nullable object that is JSON serializable.
//      Contains dynamic data, such as function arguments.
//      If null, then `[Empty]` is used instead.
//      If JSON.stringify throws an error, then `[JSON.stringify Error]` is used instead.
//
// Example where each time the tracer emits a value it'll append to the local message.
// let message = "";
// const traceManager = new TraceManager((entry: string) => (message += (entry + "\n")), "\t");
//
// function foobar(x: number): void {
//     const trace: Trace = traceManager.entry("Example", foobar.name, undefined, { x, messageLength: message.length });
//     // ...
//     trace.exit();
// }
//
// foobar(10);
// Constants used in multiple files as part of a trace's phase/task/message/details.
var TraceConstant;
(function (TraceConstant) {
    TraceConstant["CorrelationId"] = "CorrelationId";
    TraceConstant["Empty"] = "[Empty]";
    TraceConstant["Entry"] = "Entry";
    TraceConstant["Exit"] = "Exit";
    TraceConstant["IsError"] = "IsError";
    TraceConstant["IsThrowing"] = "IsThrowing";
    TraceConstant["Length"] = "Length";
    TraceConstant["Result"] = "Result";
    TraceConstant["Size"] = "Size";
})(TraceConstant = exports.TraceConstant || (exports.TraceConstant = {}));
class Trace {
    constructor(emitter, phase, task, id, correlationId, details) {
        this.emitter = emitter;
        this.phase = phase;
        this.task = task;
        this.id = id;
        this.correlationId = correlationId;
        this.timeCreated = performanceNow();
        this.entry(details);
    }
    entry(details) {
        this.trace(TraceConstant.Entry, details);
    }
    trace(message, details) {
        this.emitter(this, message, details);
    }
    exit(details) {
        this.trace(TraceConstant.Exit, details);
    }
}
exports.Trace = Trace;
// Tracing entries add the current time to its details field.
class BenchmarkTrace extends Trace {
    constructor(emitTraceFn, phase, task, id, correlationId, details) {
        super(emitTraceFn, phase, task, id, correlationId, details);
    }
    trace(message, details) {
        super.trace(message, details);
    }
}
exports.BenchmarkTrace = BenchmarkTrace;
class NoOpTrace extends Trace {
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    trace(_message, _details) { }
}
exports.NoOpTrace = NoOpTrace;
exports.NoOpTraceInstance = new NoOpTrace(
// eslint-disable-next-line @typescript-eslint/no-empty-function
(_trace, _message, _details) => { }, "", "", -1, undefined);
class TraceManager {
    constructor(valueDelimiter = ",", newline = "\r\n") {
        this.valueDelimiter = valueDelimiter;
        this.newline = newline;
        this.idFactory = createAutoIncrementIdFactory();
    }
    // Creates a new Trace instance and call its entry method.
    // Traces should be created at the start of a function, and further calls are made on Trace instance.
    entry(phase, task, correlationId, details) {
        return this.trace(phase, task, correlationId, details);
    }
    // Defaults to simple concatenation.
    formatMessage(trace, message, details) {
        const detailsJson = details !== undefined ? this.safeJsonStringify(details) : TraceConstant.Empty;
        return ([
            trace.phase,
            trace.task,
            trace.id,
            trace.correlationId,
            Math.floor(performanceNow() - trace.timeCreated),
            message,
            detailsJson,
        ].join(this.valueDelimiter) + this.newline);
    }
    // The return to the TraceManager.start function.
    // Subclass this when the TraceManager needs a different subclass of Trace.
    // Eg. BenchmarkTraceManager returns a BenchmarkTrace instance.
    trace(phase, task, correlationId, details) {
        return new Trace(this.emit.bind(this), phase, task, this.idFactory(), correlationId, details);
    }
    // Copied signature from `JSON.stringify`.
    // Subclass this by providing values for `replacer` and/or `space`.
    safeJsonStringify(obj, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    replacer, space) {
        try {
            return JSON.stringify(obj, replacer, space);
        }
        catch (_a) {
            return "[JSON.stringify Error]";
        }
    }
}
exports.TraceManager = TraceManager;
// Each trace entry gets passed to a callback function.
class ReportTraceManager extends TraceManager {
    constructor(emitter, valueDelimiter = "\t") {
        super(valueDelimiter);
        this.emitter = emitter;
    }
    emit(trace, message, details) {
        this.emitter(this.formatMessage(trace, message, details));
    }
}
exports.ReportTraceManager = ReportTraceManager;
// See BenchmarkTrace for details.
class BenchmarkTraceManager extends ReportTraceManager {
    constructor(outputFn, valueDelimiter = "\t") {
        super(outputFn, valueDelimiter);
    }
    trace(phase, task, correlationId, details) {
        return new BenchmarkTrace(this.emit.bind(this), phase, task, this.idFactory(), correlationId, details);
    }
}
exports.BenchmarkTraceManager = BenchmarkTraceManager;
// The TraceManager for DefaultSettings.
class NoOpTraceManager extends TraceManager {
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    emit(_tracer, _message, _details) { }
    trace(_phase, _task, _correlationId, _details) {
        return exports.NoOpTraceInstance;
    }
}
exports.NoOpTraceManager = NoOpTraceManager;
exports.NoOpTraceManagerInstance = new NoOpTraceManager(undefined, undefined);
function createAutoIncrementIdFactory() {
    let counter = 0;
    return () => {
        counter += 1;
        return counter;
    };
}
//# sourceMappingURL=trace.js.map