"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.normalizeIdentifier = exports.isQuotedIdentifier = exports.isRegularIdentifier = exports.isGeneralizedIdentifier = exports.getGeneralizedIdentifierLength = exports.getIdentifierLength = exports.getIdentifierKind = exports.findQuotedIdentifierQuotes = exports.IdentifierKind = void 0;
const common_1 = require("../common");
var IdentifierKind;
(function (IdentifierKind) {
    IdentifierKind["Generalized"] = "Generalized";
    IdentifierKind["Invalid"] = "Invalid";
    IdentifierKind["Quote"] = "Quote";
    IdentifierKind["QuoteRequired"] = "QuoteRequired";
    IdentifierKind["Regular"] = "Regular";
})(IdentifierKind = exports.IdentifierKind || (exports.IdentifierKind = {}));
// Assuming the text is a quoted identifier, finds the quotes that enclose the identifier.
// Otherwise returns undefined.
function findQuotedIdentifierQuotes(text, index) {
    if (text[index] !== "#") {
        return undefined;
    }
    return common_1.StringUtils.findQuotes(text, index + 1);
}
exports.findQuotedIdentifierQuotes = findQuotedIdentifierQuotes;
// Determines what kind of identifier the text is.
// It's possible that the text is a partially completed identifier,
// which is why we have the `allowTrailingPeriod` parameter.
function getIdentifierKind(text, allowTrailingPeriod) {
    if (isRegularIdentifier(text, allowTrailingPeriod)) {
        return IdentifierKind.Regular;
    }
    else if (isQuotedIdentifier(text)) {
        return isRegularIdentifier(text.slice(2, -1), false) ? IdentifierKind.Quote : IdentifierKind.QuoteRequired;
    }
    else if (isGeneralizedIdentifier(text)) {
        return IdentifierKind.Generalized;
    }
    else {
        return IdentifierKind.Invalid;
    }
}
exports.getIdentifierKind = getIdentifierKind;
// Assuming the text is an identifier, returns the length of the identifier.
function getIdentifierLength(text, index, allowTrailingPeriod) {
    const startingIndex = index;
    const textLength = text.length;
    let state = "Start" /* Start */;
    let matchLength;
    while (state !== "Done" /* Done */) {
        if (index === textLength) {
            return index - startingIndex;
        }
        switch (state) {
            case "Start" /* Start */:
                matchLength = common_1.StringUtils.regexMatchLength(common_1.Pattern.IdentifierStartCharacter, text, index);
                if (matchLength === undefined) {
                    state = "Done" /* Done */;
                }
                else {
                    state = "RegularIdentifier" /* RegularIdentifier */;
                    index += matchLength;
                }
                break;
            case "RegularIdentifier" /* RegularIdentifier */:
                // Don't consider `..` or `...` part of an identifier.
                if (allowTrailingPeriod && text[index] === "." && text[index + 1] !== ".") {
                    index += 1;
                }
                matchLength = common_1.StringUtils.regexMatchLength(common_1.Pattern.IdentifierPartCharacters, text, index);
                if (matchLength === undefined) {
                    state = "Done" /* Done */;
                }
                else {
                    index += matchLength;
                    // Don't consider `..` or `...` part of an identifier.
                    if (allowTrailingPeriod && text[index] === "." && text[index + 1] !== ".") {
                        index += 1;
                    }
                }
                break;
            default:
                throw common_1.Assert.isNever(state);
        }
    }
    return index !== startingIndex ? index - startingIndex : undefined;
}
exports.getIdentifierLength = getIdentifierLength;
// Assuming the text is a generalized identifier, returns the length of the identifier.
function getGeneralizedIdentifierLength(text, index) {
    const startingIndex = index;
    const textLength = text.length;
    let continueMatching = true;
    while (continueMatching) {
        const currentChr = text[index];
        if (currentChr === " ") {
            index += 1;
        }
        else if (currentChr === ".") {
            if (text[index - 1] === ".") {
                continueMatching = false;
                break;
            }
            index += 1;
        }
        else {
            const matchLength = common_1.StringUtils.regexMatchLength(common_1.Pattern.IdentifierPartCharacters, text, index);
            if (matchLength === undefined) {
                continueMatching = false;
                break;
            }
            index += matchLength;
        }
        if (index >= textLength) {
            continueMatching = false;
        }
    }
    return index !== startingIndex ? index - startingIndex : undefined;
}
exports.getGeneralizedIdentifierLength = getGeneralizedIdentifierLength;
function isGeneralizedIdentifier(text) {
    return getGeneralizedIdentifierLength(text, 0) === text.length;
}
exports.isGeneralizedIdentifier = isGeneralizedIdentifier;
function isRegularIdentifier(text, allowTrailingPeriod) {
    return getIdentifierLength(text, 0, allowTrailingPeriod) === text.length;
}
exports.isRegularIdentifier = isRegularIdentifier;
function isQuotedIdentifier(text) {
    return findQuotedIdentifierQuotes(text, 0) !== undefined;
}
exports.isQuotedIdentifier = isQuotedIdentifier;
// Removes the quotes from a quoted identifier if possible.
function normalizeIdentifier(text) {
    if (isQuotedIdentifier(text)) {
        const stripped = text.slice(2, -1);
        return isRegularIdentifier(stripped, false) ? stripped : text;
    }
    else {
        return text;
    }
}
exports.normalizeIdentifier = normalizeIdentifier;
//# sourceMappingURL=identifierUtils.js.map