"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.categorize = void 0;
const common_1 = require("../../../common");
const isEqualType_1 = require("./isEqualType");
const __1 = require("..");
const typeTraceConstant_1 = require("./typeTraceConstant");
// Takes a collection of PowerQueryType and breaks them down into buckets based on their TypeKind,
// then again on their ExtendedTypeKind.
function categorize(types, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.Categorize, categorize.name, correlationId);
    const categoryByKind = new Map();
    for (const type of types) {
        const category = categoryByKind.get(type.kind);
        if (category === undefined) {
            categoryByKind.set(type.kind, createCategory(type, traceManager, trace.id));
        }
        else {
            categoryByKind.set(type.kind, addToCategory(category, type, traceManager, trace.id));
        }
    }
    const result = {
        actions: categoryByKind.get(__1.Type.TypeKind.Action),
        anys: categoryByKind.get(__1.Type.TypeKind.Any),
        anyNonNulls: categoryByKind.get(__1.Type.TypeKind.AnyNonNull),
        binaries: categoryByKind.get(__1.Type.TypeKind.Binary),
        dates: categoryByKind.get(__1.Type.TypeKind.Date),
        dateTimes: categoryByKind.get(__1.Type.TypeKind.DateTime),
        dateTimeZones: categoryByKind.get(__1.Type.TypeKind.DateTimeZone),
        durations: categoryByKind.get(__1.Type.TypeKind.Duration),
        functions: categoryByKind.get(__1.Type.TypeKind.Function),
        lists: categoryByKind.get(__1.Type.TypeKind.List),
        logicals: categoryByKind.get(__1.Type.TypeKind.Logical),
        nones: categoryByKind.get(__1.Type.TypeKind.None),
        notApplicables: categoryByKind.get(__1.Type.TypeKind.NotApplicable),
        nulls: categoryByKind.get(__1.Type.TypeKind.Null),
        numbers: categoryByKind.get(__1.Type.TypeKind.Number),
        records: categoryByKind.get(__1.Type.TypeKind.Record),
        tables: categoryByKind.get(__1.Type.TypeKind.Table),
        texts: categoryByKind.get(__1.Type.TypeKind.Text),
        times: categoryByKind.get(__1.Type.TypeKind.Time),
        types: categoryByKind.get(__1.Type.TypeKind.Type),
        unknowns: categoryByKind.get(__1.Type.TypeKind.Unknown),
    };
    trace.exit();
    return result;
}
exports.categorize = categorize;
function addToCategory(category, type, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.Categorize, addToCategory.name, correlationId);
    let result;
    // We can't group cases which call `addToCategoryForPrimitive` as they each have a different generic type.
    switch (type.kind) {
        case __1.Type.TypeKind.Action:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.Any: {
            assertIsCategoryForType(category, type);
            result = addTypeIfUniqueAny(category, type);
            break;
        }
        case __1.Type.TypeKind.AnyNonNull:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.Binary:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.Date:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.DateTime:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.DateTimeZone:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.Duration:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.Function: {
            assertIsCategoryForType(category, type);
            result = addToCategoryForFunction(category, type);
            break;
        }
        case __1.Type.TypeKind.List: {
            assertIsCategoryForType(category, type);
            result = addToCategoryForList(category, type);
            break;
        }
        case __1.Type.TypeKind.Logical: {
            assertIsCategoryForType(category, type);
            result = addToCategoryForLogical(category, type);
            break;
        }
        case __1.Type.TypeKind.None:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.NotApplicable:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.Null:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.Number: {
            assertIsCategoryForType(category, type);
            result = addToCategoryForNumber(category, type);
            break;
        }
        case __1.Type.TypeKind.Record: {
            assertIsCategoryForType(category, type);
            result = addToCategoryForRecord(category, type);
            break;
        }
        case __1.Type.TypeKind.Table: {
            assertIsCategoryForType(category, type);
            result = addToCategoryForTable(category, type);
            break;
        }
        case __1.Type.TypeKind.Text: {
            assertIsCategoryForType(category, type);
            result = addToCategoryForText(category, type);
            break;
        }
        case __1.Type.TypeKind.Time:
            result = addToCategoryForPrimitive(category, type);
            break;
        case __1.Type.TypeKind.Type: {
            assertIsCategoryForType(category, type);
            result = addTypeIfUniqueType(category, type);
            break;
        }
        case __1.Type.TypeKind.Unknown:
            result = addToCategoryForPrimitive(category, type);
            break;
        default:
            throw common_1.Assert.isNever(type);
    }
    trace.exit();
    return result;
}
function addTypeIfUniqueAny(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.AnyUnion:
            return Object.assign(Object.assign({}, category), { anyUnions: category.anyUnions.add(type), flattenedAnyUnions: category.flattenedAnyUnions.addMany(flattenAnyUnion(type)) });
        case undefined:
            return addToCategoryForPrimitive(category, type);
        default:
            throw common_1.Assert.isNever(type);
    }
}
function addToCategoryForFunction(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedFunction: {
            return Object.assign(Object.assign({}, category), { definedFunctions: category.definedFunctions.add(type) });
        }
        case undefined: {
            return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function addToCategoryForList(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedList: {
            return Object.assign(Object.assign({}, category), { definedLists: category.definedLists.add(type) });
        }
        case undefined: {
            return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function addToCategoryForLogical(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.LogicalLiteral: {
            return Object.assign(Object.assign({}, category), { hasFalsyNonNullableLiteral: category.hasFalsyNonNullableLiteral || (!type.normalizedLiteral && !type.isNullable), hasFalsyNullableLiteral: category.hasFalsyNullableLiteral || (!type.normalizedLiteral && type.isNullable), hasTruthyNonNullableLiteral: category.hasTruthyNonNullableLiteral || (type.normalizedLiteral && !type.isNullable), hasTruthyNullableLiteral: category.hasTruthyNullableLiteral || (type.normalizedLiteral && type.isNullable) });
        }
        case undefined: {
            return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function addToCategoryForNumber(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.NumberLiteral: {
            return Object.assign(Object.assign({}, category), { literals: category.literals.add(type) });
        }
        case undefined: {
            return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function addToCategoryForPrimitive(category, type) {
    assertIsCategoryForType(category, type);
    return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
}
function addToCategoryForRecord(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedRecord: {
            return Object.assign(Object.assign({}, category), { definedRecords: category.definedRecords.add(type) });
        }
        case undefined: {
            return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function addToCategoryForTable(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedTable: {
            return Object.assign(Object.assign({}, category), { definedTables: category.definedTables.add(type) });
        }
        case undefined: {
            return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function addToCategoryForText(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.TextLiteral: {
            return Object.assign(Object.assign({}, category), { literals: category.literals.add(type) });
        }
        case undefined: {
            return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function addTypeIfUniqueType(category, type) {
    assertIsCategoryForType(category, type);
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedListType: {
            return Object.assign(Object.assign({}, category), { definedListTypes: category.definedListTypes.add(type) });
        }
        case __1.Type.ExtendedTypeKind.FunctionType: {
            return Object.assign(Object.assign({}, category), { functionTypes: category.functionTypes.add(type) });
        }
        case __1.Type.ExtendedTypeKind.ListType: {
            return Object.assign(Object.assign({}, category), { listTypes: category.listTypes.add(type) });
        }
        case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType: {
            return Object.assign(Object.assign({}, category), { primaryPrimitiveTypes: category.primaryPrimitiveTypes.add(type) });
        }
        case __1.Type.ExtendedTypeKind.RecordType: {
            return Object.assign(Object.assign({}, category), { recordTypes: category.recordTypes.add(type) });
        }
        case __1.Type.ExtendedTypeKind.TableType: {
            return Object.assign(Object.assign({}, category), { tableTypes: category.tableTypes.add(type) });
        }
        case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression: {
            return Object.assign(Object.assign({}, category), { tablePrimaryExpressionTypes: category.tablePrimaryExpressionTypes.add(type) });
        }
        case undefined: {
            return Object.assign(Object.assign({}, category), { primitives: category.primitives.add(type) });
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function assertIsCategoryForType(category, type) {
    if (category.kind !== type.kind) {
        throw new common_1.CommonError.InvariantError(`expected category and type to have the same kind`, {
            categoryKind: category.kind,
            typeKind: type.kind,
        });
    }
}
function createCategory(type, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.Categorize, createCategory.name, correlationId);
    let result;
    switch (type.kind) {
        case __1.Type.TypeKind.Action:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.Any:
            result = createCategoryForAny(type);
            break;
        case __1.Type.TypeKind.AnyNonNull:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.Binary:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.Date:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.DateTime:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.DateTimeZone:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.Duration:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.Function:
            result = createCategoryForFunction(type);
            break;
        case __1.Type.TypeKind.List:
            result = createCategoryForList(type);
            break;
        case __1.Type.TypeKind.Logical:
            result = createCategoryForLogical(type);
            break;
        case __1.Type.TypeKind.None:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.NotApplicable:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.Null:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.Number:
            result = createCategoryForNumber(type);
            break;
        case __1.Type.TypeKind.Record:
            result = createCategoryForRecord(type);
            break;
        case __1.Type.TypeKind.Table:
            result = createCategoryForTable(type);
            break;
        case __1.Type.TypeKind.Text:
            result = createCategoryForText(type);
            break;
        case __1.Type.TypeKind.Time:
            result = createCategoryForPrimitive(type);
            break;
        case __1.Type.TypeKind.Type:
            result = createCategoryForType(type);
            break;
        case __1.Type.TypeKind.Unknown:
            result = createCategoryForPrimitive(type);
            break;
        default:
            throw common_1.Assert.isNever(type);
    }
    trace.exit();
    return result;
}
function createCategoryForAny(type) {
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.AnyUnion: {
            return {
                kind: __1.Type.TypeKind.Any,
                primitives: new common_1.ImmutableSet([], isEqualType_1.isEqualPrimitiveType),
                anyUnions: new common_1.ImmutableSet([type], isEqualType_1.isEqualAnyUnion),
                flattenedAnyUnions: new common_1.ImmutableSet(flattenAnyUnion(type), isEqualType_1.isEqualType),
            };
        }
        case undefined: {
            return {
                kind: __1.Type.TypeKind.Any,
                primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualPrimitiveType),
                anyUnions: new common_1.ImmutableSet([], isEqualType_1.isEqualAnyUnion),
                flattenedAnyUnions: new common_1.ImmutableSet([], isEqualType_1.isEqualType),
            };
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function createCategoryForFunction(type) {
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedFunction: {
            return {
                kind: __1.Type.TypeKind.Function,
                primitives: new common_1.ImmutableSet([], isEqualType_1.isEqualPrimitiveType),
                definedFunctions: new common_1.ImmutableSet([type], isEqualType_1.isEqualDefinedFunction),
            };
        }
        case undefined: {
            return {
                kind: __1.Type.TypeKind.Function,
                primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualPrimitiveType),
                definedFunctions: new common_1.ImmutableSet([], isEqualType_1.isEqualDefinedFunction),
            };
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function createCategoryForList(type) {
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedList: {
            return {
                kind: __1.Type.TypeKind.List,
                primitives: new common_1.ImmutableSet([], isEqualType_1.isEqualPrimitiveType),
                definedLists: new common_1.ImmutableSet([type], isEqualType_1.isEqualDefinedList),
            };
        }
        case undefined: {
            return {
                kind: __1.Type.TypeKind.List,
                primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualPrimitiveType),
                definedLists: new common_1.ImmutableSet([], isEqualType_1.isEqualDefinedList),
            };
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function createCategoryForLogical(type) {
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.LogicalLiteral: {
            return {
                kind: __1.Type.TypeKind.Logical,
                primitives: new common_1.ImmutableSet([], isEqualType_1.isEqualPrimitiveType),
                hasFalsyNonNullableLiteral: !type.normalizedLiteral && !type.isNullable,
                hasFalsyNullableLiteral: !type.normalizedLiteral && type.isNullable,
                hasTruthyNonNullableLiteral: type.normalizedLiteral && !type.isNullable,
                hasTruthyNullableLiteral: type.normalizedLiteral && type.isNullable,
            };
        }
        case undefined: {
            return {
                kind: __1.Type.TypeKind.Logical,
                primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualPrimitiveType),
                hasFalsyNonNullableLiteral: false,
                hasFalsyNullableLiteral: false,
                hasTruthyNonNullableLiteral: false,
                hasTruthyNullableLiteral: false,
            };
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function createCategoryForNumber(type) {
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.NumberLiteral: {
            return {
                kind: __1.Type.TypeKind.Number,
                primitives: new common_1.ImmutableSet([], isEqualType_1.isEqualPrimitiveType),
                literals: new common_1.ImmutableSet([type], isEqualType_1.isEqualNumberLiteral),
            };
        }
        case undefined: {
            return {
                kind: __1.Type.TypeKind.Number,
                primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualPrimitiveType),
                literals: new common_1.ImmutableSet([], isEqualType_1.isEqualNumberLiteral),
            };
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function createCategoryForPrimitive(type) {
    return {
        kind: type.kind,
        primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualType),
    };
}
function createCategoryForRecord(type) {
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedRecord: {
            return {
                kind: __1.Type.TypeKind.Record,
                primitives: new common_1.ImmutableSet([], isEqualType_1.isEqualPrimitiveType),
                definedRecords: new common_1.ImmutableSet([type], isEqualType_1.isEqualType),
            };
        }
        case undefined: {
            return {
                kind: __1.Type.TypeKind.Record,
                primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualPrimitiveType),
                definedRecords: new common_1.ImmutableSet([], isEqualType_1.isEqualDefinedRecord),
            };
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function createCategoryForTable(type) {
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedTable: {
            return {
                kind: __1.Type.TypeKind.Table,
                primitives: new common_1.ImmutableSet([], isEqualType_1.isEqualPrimitiveType),
                definedTables: new common_1.ImmutableSet([type], isEqualType_1.isEqualDefinedTable),
            };
        }
        case undefined: {
            return {
                kind: __1.Type.TypeKind.Table,
                primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualPrimitiveType),
                definedTables: new common_1.ImmutableSet([], isEqualType_1.isEqualDefinedTable),
            };
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function createCategoryForText(type) {
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.TextLiteral: {
            return {
                kind: __1.Type.TypeKind.Text,
                primitives: new common_1.ImmutableSet([], isEqualType_1.isEqualPrimitiveType),
                literals: new common_1.ImmutableSet([type], isEqualType_1.isEqualTextLiteral),
            };
        }
        case undefined: {
            return {
                kind: __1.Type.TypeKind.Text,
                primitives: new common_1.ImmutableSet([type], isEqualType_1.isEqualPrimitiveType),
                literals: new common_1.ImmutableSet([], isEqualType_1.isEqualTextLiteral),
            };
        }
        default:
            throw common_1.Assert.isNever(type);
    }
}
function createCategoryForType(type) {
    let primitives = [];
    let definedListTypes = [];
    let functionTypes = [];
    let listTypes = [];
    let primaryPrimitiveTypes = [];
    let recordTypes = [];
    let tablePrimaryExpressionTypes = [];
    let tableTypes = [];
    switch (type.extendedKind) {
        case __1.Type.ExtendedTypeKind.DefinedListType:
            definedListTypes = [type];
            break;
        case __1.Type.ExtendedTypeKind.FunctionType:
            functionTypes = [type];
            break;
        case __1.Type.ExtendedTypeKind.ListType:
            listTypes = [type];
            break;
        case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType:
            primaryPrimitiveTypes = [type];
            break;
        case __1.Type.ExtendedTypeKind.RecordType:
            recordTypes = [type];
            break;
        case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression:
            tablePrimaryExpressionTypes = [type];
            break;
        case __1.Type.ExtendedTypeKind.TableType:
            tableTypes = [type];
            break;
        case undefined: {
            primitives = [type];
            break;
        }
        default:
            throw common_1.Assert.isNever(type);
    }
    return {
        kind: __1.Type.TypeKind.Type,
        primitives: new common_1.ImmutableSet(primitives, isEqualType_1.isEqualPrimitiveType),
        definedListTypes: new common_1.ImmutableSet(definedListTypes, isEqualType_1.isEqualDefinedListType),
        functionTypes: new common_1.ImmutableSet(functionTypes, isEqualType_1.isEqualFunctionType),
        listTypes: new common_1.ImmutableSet(listTypes, isEqualType_1.isEqualListType),
        primaryPrimitiveTypes: new common_1.ImmutableSet(primaryPrimitiveTypes, isEqualType_1.isEqualPrimaryPrimitiveType),
        recordTypes: new common_1.ImmutableSet(recordTypes, isEqualType_1.isEqualRecordType),
        tablePrimaryExpressionTypes: new common_1.ImmutableSet(tablePrimaryExpressionTypes, isEqualType_1.isEqualTableTypePrimaryExpression),
        tableTypes: new common_1.ImmutableSet(tableTypes, isEqualType_1.isEqualTableType),
    };
}
function flattenAnyUnion(anyUnion) {
    let newUnionedTypePairs = [];
    for (const item of anyUnion.unionedTypePairs) {
        // If it's an Any primitive then we can do an early return.
        // Else it's an AnyUnion so continue flattening the types.
        if (item.kind === __1.Type.TypeKind.Any) {
            switch (item.extendedKind) {
                case undefined:
                    return [item];
                case __1.Type.ExtendedTypeKind.AnyUnion:
                    newUnionedTypePairs = newUnionedTypePairs.concat(flattenAnyUnion(item));
                    break;
                default:
                    throw common_1.Assert.isNever(item);
            }
        }
        else {
            newUnionedTypePairs.push(item);
        }
    }
    return newUnionedTypePairs;
}
//# sourceMappingURL=categorize.js.map