"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.isCompatibleWithFunctionParameter = exports.isCompatibleWithFunctionSignature = exports.isCompatible = void 0;
const common_1 = require("../../../common");
const isEqualType_1 = require("./isEqualType");
const isType_1 = require("./isType");
const __1 = require("..");
const typeTraceConstant_1 = require("./typeTraceConstant");
// Returns `${left} is compatible with ${right}. Eg.
// `Type.TextInstance is compatible with Type.AnyInstance` -> true
// `Type.AnyInstance is compatible with Type.TextInstance` -> false
// `Type.NullInstance is compatible with Type.AnyNonNull` -> false
// `Type.TextInstance is compatible with Type.AnyUnion([Type.TextInstance, Type.NumberInstance])` -> true
function isCompatible(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatible.name, correlationId);
    let result;
    if (left.kind === __1.Type.TypeKind.NotApplicable ||
        left.kind === __1.Type.TypeKind.Unknown ||
        right.kind === __1.Type.TypeKind.NotApplicable ||
        right.kind === __1.Type.TypeKind.Unknown) {
        result = undefined;
    }
    else if (left.kind === __1.Type.TypeKind.None ||
        right.kind === __1.Type.TypeKind.None ||
        (left.isNullable && !right.isNullable)) {
        result = false;
    }
    else if (left.kind === __1.Type.TypeKind.Null && right.isNullable) {
        result = true;
    }
    else {
        switch (right.kind) {
            case __1.Type.TypeKind.Action:
            case __1.Type.TypeKind.Binary:
            case __1.Type.TypeKind.Date:
            case __1.Type.TypeKind.DateTime:
            case __1.Type.TypeKind.DateTimeZone:
            case __1.Type.TypeKind.Duration:
            case __1.Type.TypeKind.Time:
                result = (0, isEqualType_1.isEqualType)(left, right);
                break;
            case __1.Type.TypeKind.Any:
                result = isCompatibleWithAny(left, right, traceManager, trace.id);
                break;
            case __1.Type.TypeKind.AnyNonNull:
                result = left.kind !== __1.Type.TypeKind.Null && !left.isNullable;
                break;
            case __1.Type.TypeKind.Function:
                result = isCompatibleWithFunction(left, right, traceManager, trace.id);
                break;
            case __1.Type.TypeKind.List:
                result = isCompatibleWithList(left, right, traceManager, trace.id);
                break;
            case __1.Type.TypeKind.Logical:
                result = isCompatibleWithPrimitiveOrLiteral(left, right);
                break;
            case __1.Type.TypeKind.Number:
                result = isCompatibleWithPrimitiveOrLiteral(left, right);
                break;
            case __1.Type.TypeKind.Null:
                result = left.kind === __1.Type.TypeKind.Null;
                break;
            case __1.Type.TypeKind.Record:
                result = isCompatibleWithRecord(left, right, traceManager, trace.id);
                break;
            case __1.Type.TypeKind.Table:
                result = isCompatibleWithTable(left, right, traceManager, trace.id);
                break;
            case __1.Type.TypeKind.Text:
                result = isCompatibleWithPrimitiveOrLiteral(left, right);
                break;
            case __1.Type.TypeKind.Type:
                result = isCompatibleWithType(left, right, traceManager, trace.id);
                break;
            default:
                throw common_1.Assert.isNever(right);
        }
    }
    trace.exit();
    return result;
}
exports.isCompatible = isCompatible;
function isCompatibleWithFunctionSignature(left, right) {
    if (!(0, isType_1.isFunctionSignature)(left)) {
        return false;
    }
    return (0, isEqualType_1.isEqualFunctionSignature)(left, right);
}
exports.isCompatibleWithFunctionSignature = isCompatibleWithFunctionSignature;
function isCompatibleWithFunctionParameter(left, right) {
    if (left === undefined) {
        return right.isOptional;
    }
    else if (left.isNullable && !right.isNullable) {
        return false;
    }
    else {
        return (!right.type ||
            right.type === __1.Type.TypeKind.Any ||
            left.kind === __1.Type.TypeKind.Any ||
            (left.kind === __1.Type.TypeKind.Null && right.isNullable) ||
            left.kind === right.type);
    }
}
exports.isCompatibleWithFunctionParameter = isCompatibleWithFunctionParameter;
function isCompatibleWithAny(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithAny.name, correlationId);
    let result;
    switch (right.extendedKind) {
        case undefined:
            result = true;
            break;
        case __1.Type.ExtendedTypeKind.AnyUnion:
            result = isCompatibleWithAnyUnion(left, right, traceManager, trace.id);
            break;
        default:
            throw common_1.Assert.isNever(right);
    }
    trace.exit();
    return result;
}
function isCompatibleWithAnyUnion(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithAnyUnion.name, correlationId);
    for (const subtype of right.unionedTypePairs) {
        if (isCompatible(left, subtype, traceManager, trace.id)) {
            trace.exit();
            return true;
        }
    }
    trace.exit();
    return false;
}
function isCompatibleWithDefinedList(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithDefinedList.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (left.extendedKind) {
            case undefined:
                result = false;
                break;
            case __1.Type.ExtendedTypeKind.DefinedList: {
                result = isCompatibleDefinedListOrDefinedListType(left, right, traceManager, trace.id);
                break;
            }
            default:
                throw common_1.Assert.isNever(left);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithDefinedListType(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithDefinedListType.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (left.extendedKind) {
            case __1.Type.ExtendedTypeKind.DefinedListType:
                result = isCompatibleDefinedListOrDefinedListType(left, right, traceManager, trace.id);
                break;
            case __1.Type.ExtendedTypeKind.ListType:
                result = isDefinedListTypeCompatibleWithListType(right, left, traceManager, trace.id);
                break;
            case undefined:
            case __1.Type.ExtendedTypeKind.FunctionType:
            case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType:
            case __1.Type.ExtendedTypeKind.RecordType:
            case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression:
            case __1.Type.ExtendedTypeKind.TableType:
                result = false;
                break;
            default:
                throw common_1.Assert.isNever(left);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithDefinedRecord(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithDefinedRecord.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (left.extendedKind) {
            case undefined:
                result = false;
                break;
            case __1.Type.ExtendedTypeKind.DefinedRecord:
                result = isCompatibleWithFieldSpecificationList(left, right, traceManager, trace.id);
                break;
            default:
                throw common_1.Assert.isNever(left);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithDefinedTable(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithDefinedTable.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (left.extendedKind) {
            case undefined:
                result = false;
                break;
            case __1.Type.ExtendedTypeKind.DefinedTable: {
                result = isCompatibleWithFieldSpecificationList(left, right, traceManager, trace.id);
                break;
            }
            default:
                throw common_1.Assert.isNever(left);
        }
    }
    trace.exit();
    return result;
}
// TODO: decide what a compatible FieldSpecificationList should look like
function isCompatibleWithFieldSpecificationList(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithFieldSpecificationList.name, correlationId);
    let result;
    if (!(0, isType_1.isFieldSpecificationList)(left)) {
        result = false;
    }
    else {
        result = common_1.MapUtils.isSubsetMap(left.fields, right.fields, (leftValue, rightValue) => {
            const subsetResult = isCompatible(leftValue, rightValue, traceManager, trace.id);
            return subsetResult !== undefined && subsetResult;
        });
    }
    trace.exit();
    return result;
}
function isCompatibleWithFunction(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithFieldSpecificationList.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (right.extendedKind) {
            case undefined:
                result = true;
                break;
            case __1.Type.ExtendedTypeKind.DefinedFunction:
                result = isCompatibleWithFunctionSignature(left, right);
                break;
            default:
                throw common_1.Assert.isNever(right);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithList(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithList.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (right.extendedKind) {
            case undefined:
                result = left.extendedKind === undefined;
                break;
            case __1.Type.ExtendedTypeKind.DefinedList:
                result = isCompatibleWithDefinedList(left, right, traceManager, trace.id);
                break;
            default:
                throw common_1.Assert.isNever(right);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithListType(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithListType.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (left.extendedKind) {
            case __1.Type.ExtendedTypeKind.DefinedListType:
                result = isDefinedListTypeCompatibleWithListType(left, right, traceManager, trace.id);
                break;
            case __1.Type.ExtendedTypeKind.ListType:
                result = (0, isEqualType_1.isEqualType)(left.itemType, right.itemType);
                break;
            case undefined:
            case __1.Type.ExtendedTypeKind.FunctionType:
            case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType:
            case __1.Type.ExtendedTypeKind.RecordType:
            case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression:
            case __1.Type.ExtendedTypeKind.TableType:
                result = false;
                break;
            default:
                throw common_1.Assert.isNever(left);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithPrimaryPrimitiveType(left, right) {
    if (left.kind !== right.kind) {
        return false;
    }
    switch (left.extendedKind) {
        case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType:
            return left.primitiveType === right.primitiveType;
        case undefined:
        case __1.Type.ExtendedTypeKind.DefinedListType:
        case __1.Type.ExtendedTypeKind.ListType:
        case __1.Type.ExtendedTypeKind.FunctionType:
        case __1.Type.ExtendedTypeKind.RecordType:
        case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression:
        case __1.Type.ExtendedTypeKind.TableType:
            return false;
        default:
            throw common_1.Assert.isNever(left);
    }
}
function isCompatibleWithRecord(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithRecord.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (right.extendedKind) {
            case undefined:
                result = true;
                break;
            case __1.Type.ExtendedTypeKind.DefinedRecord:
                result = isCompatibleWithDefinedRecord(left, right, traceManager, trace.id);
                break;
            default:
                throw common_1.Assert.isNever(right);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithRecordType(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithRecordType.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (left.extendedKind) {
            case __1.Type.ExtendedTypeKind.RecordType:
                result = isCompatibleWithFieldSpecificationList(left, right, traceManager, trace.id);
                break;
            case undefined:
            case __1.Type.ExtendedTypeKind.DefinedListType:
            case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType:
            case __1.Type.ExtendedTypeKind.ListType:
            case __1.Type.ExtendedTypeKind.FunctionType:
            case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression:
            case __1.Type.ExtendedTypeKind.TableType:
                result = false;
                break;
            default:
                throw common_1.Assert.isNever(left);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithTable(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithTable.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (right.extendedKind) {
            case undefined:
                result = true;
                break;
            case __1.Type.ExtendedTypeKind.DefinedTable:
                result = isCompatibleWithDefinedTable(left, right, traceManager, trace.id);
                break;
            default:
                throw common_1.Assert.isNever(right);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithTableType(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithTableType.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (left.extendedKind) {
            case undefined:
                result = false;
                break;
            case __1.Type.ExtendedTypeKind.TableType:
                result = isCompatibleWithFieldSpecificationList(left, right, traceManager, trace.id);
                break;
            case __1.Type.ExtendedTypeKind.DefinedListType:
            case __1.Type.ExtendedTypeKind.ListType:
            case __1.Type.ExtendedTypeKind.FunctionType:
            case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType:
            case __1.Type.ExtendedTypeKind.RecordType:
            case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression:
                result = false;
                break;
            default:
                throw common_1.Assert.isNever(left);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithTableTypePrimaryExpression(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithTableTypePrimaryExpression.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (left.extendedKind) {
            case undefined:
                result = false;
                break;
            case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression:
                result = isCompatible(left.primaryExpression, right.primaryExpression, traceManager, correlationId);
                break;
            case __1.Type.ExtendedTypeKind.DefinedListType:
            case __1.Type.ExtendedTypeKind.ListType:
            case __1.Type.ExtendedTypeKind.FunctionType:
            case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType:
            case __1.Type.ExtendedTypeKind.RecordType:
            case __1.Type.ExtendedTypeKind.TableType:
                result = false;
                break;
            default:
                throw common_1.Assert.isNever(left);
        }
    }
    trace.exit();
    return result;
}
function isCompatibleWithLiteral(left, right) {
    if (left.kind !== right.kind || !left.extendedKind || left.extendedKind !== right.extendedKind) {
        return false;
    }
    else {
        return left.normalizedLiteral === right.normalizedLiteral;
    }
}
function isCompatibleDefinedListOrDefinedListType(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleDefinedListOrDefinedListType.name, correlationId);
    let leftElements;
    let rightElements;
    if (left.extendedKind === __1.Type.ExtendedTypeKind.DefinedList &&
        right.extendedKind === __1.Type.ExtendedTypeKind.DefinedList) {
        leftElements = left.elements;
        rightElements = right.elements;
    }
    else if (left.extendedKind === __1.Type.ExtendedTypeKind.DefinedListType &&
        right.extendedKind === __1.Type.ExtendedTypeKind.DefinedListType) {
        leftElements = left.itemTypes;
        rightElements = right.itemTypes;
    }
    else {
        throw new common_1.CommonError.InvariantError(`unknown scenario for isCompatibleDefinedListOrDefinedListType`, {
            leftTypeKind: left.kind,
            rightTypeKind: right.kind,
            leftExtendedTypeKind: left.extendedKind,
            rightExtendedTypeKind: right.extendedKind,
        });
    }
    if (leftElements.length !== rightElements.length) {
        trace.exit();
        return false;
    }
    const numElements = leftElements.length;
    for (let index = 0; index < numElements; index += 1) {
        if (!isCompatible(leftElements[index], rightElements[index], traceManager, trace.id)) {
            trace.exit();
            return false;
        }
    }
    trace.exit();
    return true;
}
function isCompatibleWithPrimitiveOrLiteral(left, right) {
    return left.kind === right.kind && (!right.extendedKind || isCompatibleWithLiteral(left, right));
}
function isCompatibleWithType(left, right, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isCompatibleWithType.name, correlationId);
    let result;
    if (left.kind !== right.kind) {
        result = false;
    }
    else {
        switch (right.extendedKind) {
            case undefined:
                result = true;
                break;
            case __1.Type.ExtendedTypeKind.FunctionType:
                result = isCompatibleWithFunctionSignature(left, right);
                break;
            case __1.Type.ExtendedTypeKind.ListType:
                result = isCompatibleWithListType(left, right, traceManager, trace.id);
                break;
            case __1.Type.ExtendedTypeKind.DefinedListType:
                result = isCompatibleWithDefinedListType(left, right, traceManager, trace.id);
                break;
            case __1.Type.ExtendedTypeKind.PrimaryPrimitiveType:
                result = isCompatibleWithPrimaryPrimitiveType(left, right);
                break;
            case __1.Type.ExtendedTypeKind.RecordType:
                result = isCompatibleWithRecordType(left, right, traceManager, trace.id);
                break;
            case __1.Type.ExtendedTypeKind.TableType:
                result = isCompatibleWithTableType(left, right, traceManager, trace.id);
                break;
            case __1.Type.ExtendedTypeKind.TableTypePrimaryExpression:
                result = isCompatibleWithTableTypePrimaryExpression(left, right, traceManager, trace.id);
                break;
            default:
                throw common_1.Assert.isNever(right);
        }
    }
    trace.exit();
    return result;
}
function isDefinedListTypeCompatibleWithListType(definedList, listType, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.IsCompatible, isDefinedListTypeCompatibleWithListType.name, correlationId);
    const itemTypeCompatabilities = definedList.itemTypes.map((itemType) => isCompatible(itemType, listType.itemType, traceManager, trace.id));
    const result = Boolean(itemTypeCompatabilities.find((value) => value === undefined || value === false));
    trace.exit();
    return result;
}
//# sourceMappingURL=isCompatible.js.map