"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.nameOfTypeKind = exports.nameOfFunctionSignature = exports.nameOfFunctionParameter = exports.nameOf = void 0;
const common_1 = require("../../../common");
const __1 = require("../..");
const __2 = require("..");
const typeTraceConstant_1 = require("./typeTraceConstant");
function nameOf(type, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.NameOf, nameOf.name, correlationId);
    let result;
    if (type.kind === __2.Type.TypeKind.Null) {
        result = "null";
    }
    else {
        switch (type.extendedKind) {
            case __2.Type.ExtendedTypeKind.NumberLiteral:
            case __2.Type.ExtendedTypeKind.TextLiteral:
                result = prefixNullableIfRequired(type, `${type.literal}`);
                break;
            case __2.Type.ExtendedTypeKind.AnyUnion:
                result = type.unionedTypePairs
                    .map((subtype) => nameOf(subtype, traceManager, trace.id))
                    .join(" | ");
                break;
            case __2.Type.ExtendedTypeKind.DefinedFunction:
                result = prefixNullableIfRequired(type, nameOfFunctionSignature(type, true, traceManager, trace.id));
                break;
            case __2.Type.ExtendedTypeKind.DefinedList:
                result = prefixNullableIfRequired(type, `{${nameOfIterable(type.elements, traceManager, trace.id)}}`);
                break;
            case __2.Type.ExtendedTypeKind.DefinedListType:
                result = prefixNullableIfRequired(type, `type {${nameOfIterable(type.itemTypes, traceManager, trace.id)}}`);
                break;
            case __2.Type.ExtendedTypeKind.DefinedRecord:
                result = prefixNullableIfRequired(type, nameOfFieldSpecificationList(type, traceManager, trace.id));
                break;
            case __2.Type.ExtendedTypeKind.DefinedTable:
                result = prefixNullableIfRequired(type, `table ${nameOfFieldSpecificationList(type, traceManager, trace.id)}`);
                break;
            case __2.Type.ExtendedTypeKind.FunctionType:
                result = prefixNullableIfRequired(type, `type function ${nameOfFunctionSignature(type, false, traceManager, trace.id)}`);
                break;
            case __2.Type.ExtendedTypeKind.ListType:
                result = prefixNullableIfRequired(type, `type {${nameOf(type.itemType, traceManager, trace.id)}}`);
                break;
            case __2.Type.ExtendedTypeKind.LogicalLiteral:
                result = prefixNullableIfRequired(type, `${type.normalizedLiteral}`);
                break;
            case __2.Type.ExtendedTypeKind.PrimaryPrimitiveType:
                result = prefixNullableIfRequired(type, `type ${nameOf(type.primitiveType, traceManager, trace.id)}`);
                break;
            case __2.Type.ExtendedTypeKind.RecordType:
                result = prefixNullableIfRequired(type, `type ${nameOfFieldSpecificationList(type, traceManager, trace.id)}`);
                break;
            case __2.Type.ExtendedTypeKind.TableType:
                result = prefixNullableIfRequired(type, `type table ${nameOfFieldSpecificationList(type, traceManager, trace.id)}`);
                break;
            case __2.Type.ExtendedTypeKind.TableTypePrimaryExpression:
                result = prefixNullableIfRequired(type, `type table ${nameOf(type.primaryExpression, traceManager, trace.id)}`);
                break;
            case undefined:
                result = prefixNullableIfRequired(type, nameOfTypeKind(type.kind));
                break;
            default:
                throw common_1.Assert.isNever(type);
        }
    }
    trace.exit();
    return result;
}
exports.nameOf = nameOf;
function nameOfFunctionParameter(parameter, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.NameOf, nameOfFunctionParameter.name, correlationId);
    let partial = `${parameter.nameLiteral}:`;
    if (parameter.isOptional) {
        partial += " optional";
    }
    if (parameter.isNullable) {
        partial += " nullable";
    }
    if (parameter.type !== undefined) {
        partial += ` ${nameOfTypeKind(parameter.type)}`;
    }
    else {
        partial += ` ${nameOfTypeKind(__2.Type.TypeKind.Any)}`;
    }
    trace.exit();
    return partial;
}
exports.nameOfFunctionParameter = nameOfFunctionParameter;
function nameOfFunctionSignature(type, includeFatArrow, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.NameOf, nameOfFunctionSignature.name, correlationId);
    const parameters = type.parameters
        .map((parameter) => nameOfFunctionParameter(parameter, traceManager, trace.id))
        .join(", ");
    const result = `(${parameters})${includeFatArrow ? " =>" : ""} ${nameOf(type.returnType, traceManager, trace.id)}`;
    trace.exit();
    return result;
}
exports.nameOfFunctionSignature = nameOfFunctionSignature;
function nameOfTypeKind(kind) {
    return kind === __2.Type.TypeKind.NotApplicable ? "not applicable" : kind.toLowerCase();
}
exports.nameOfTypeKind = nameOfTypeKind;
function nameOfFieldSpecificationList(type, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.NameOf, nameOfFieldSpecificationList.name, correlationId);
    const chunks = [];
    for (const [key, value] of type.fields.entries()) {
        chunks.push(`${key}: ${nameOf(value, traceManager, trace.id)}`);
    }
    if (type.isOpen) {
        chunks.push("...");
    }
    const pairs = chunks.join(", ");
    const result = `[${pairs}]`;
    trace.exit();
    return result;
}
function nameOfIterable(collection, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.NameOf, nameOfIterable.name, correlationId);
    const result = collection
        .map((item) => nameOf(item, traceManager, trace.id))
        .join(", ");
    trace.exit();
    return result;
}
function prefixNullableIfRequired(type, name) {
    return type.isNullable ? `${__1.Constant.LanguageConstant.Nullable} ${name}` : name;
}
//# sourceMappingURL=nameOf.js.map