"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.typeCheckTable = exports.typeCheckRecord = exports.typeCheckListWithDefinedListType = exports.typeCheckListWithListType = exports.typeCheckInvocation = exports.typeCheckFunctionSignature = exports.typeCheckFunction = void 0;
const isCompatible_1 = require("./isCompatible");
const common_1 = require("../../../common");
const isEqualType_1 = require("./isEqualType");
const typeTraceConstant_1 = require("./typeTraceConstant");
function typeCheckFunction(valueType, schemaType, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.TypeCheck, typeCheckFunction.name, correlationId);
    const result = Object.assign(Object.assign({}, typeCheckFunctionSignature(valueType, schemaType, traceManager, trace.id)), { isReturnTypeCompatible: (0, isCompatible_1.isCompatible)(valueType.returnType, schemaType.returnType, traceManager, trace.id) === true });
    trace.exit();
    return result;
}
exports.typeCheckFunction = typeCheckFunction;
function typeCheckFunctionSignature(valueType, schemaType, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.TypeCheck, typeCheckFunctionSignature.name, correlationId);
    const result = typeCheckGenericNumber(valueType.parameters, schemaType.parameters, (left, right) => (0, isEqualType_1.isEqualFunctionParameter)(left, right), traceManager, trace.id);
    trace.exit();
    return result;
}
exports.typeCheckFunctionSignature = typeCheckFunctionSignature;
function typeCheckInvocation(args, definedFunction, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.TypeCheck, typeCheckInvocation.name, correlationId);
    const parameters = definedFunction.parameters;
    const numArgs = args.length;
    const numParameters = parameters.length;
    const extraneousArgs = numArgs > numParameters ? common_1.ArrayUtils.range(numArgs - numParameters, numParameters) : [];
    const validArgs = [];
    const missingArgs = [];
    const invalidArgs = new Map();
    for (let index = 0; index < numParameters; index += 1) {
        const arg = args[index];
        const parameter = parameters[index];
        if ((0, isCompatible_1.isCompatibleWithFunctionParameter)(arg, parameter)) {
            validArgs.push(index);
        }
        else if (arg !== undefined) {
            invalidArgs.set(index, {
                expected: parameter,
                actual: arg,
            });
        }
        else {
            missingArgs.push(index);
        }
    }
    const result = {
        valid: validArgs,
        invalid: invalidArgs,
        extraneous: extraneousArgs,
        missing: missingArgs,
    };
    trace.exit();
    return result;
}
exports.typeCheckInvocation = typeCheckInvocation;
function typeCheckListWithListType(valueType, schemaType, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.TypeCheck, typeCheckListWithListType.name, correlationId);
    const validArgs = [];
    const invalidArgs = new Map();
    const schemaItemType = schemaType.itemType;
    const valueElements = valueType.elements;
    for (const [element, index] of common_1.ArrayUtils.enumerate(valueElements)) {
        if ((0, isCompatible_1.isCompatible)(element, schemaItemType, traceManager, trace.id)) {
            validArgs.push(index);
        }
        else {
            invalidArgs.set(index, {
                expected: schemaType,
                actual: element,
            });
        }
    }
    const result = {
        valid: validArgs,
        invalid: invalidArgs,
        extraneous: [],
        missing: [],
    };
    trace.exit();
    return result;
}
exports.typeCheckListWithListType = typeCheckListWithListType;
function typeCheckListWithDefinedListType(valueType, schemaType, traceManager, correlationId) {
    return typeCheckGenericNumber(valueType.elements, schemaType.itemTypes, isCompatible_1.isCompatible, traceManager, correlationId);
}
exports.typeCheckListWithDefinedListType = typeCheckListWithDefinedListType;
function typeCheckRecord(valueType, schemaType, traceManager, correlationId) {
    return typeCheckRecordOrTable(valueType.fields, schemaType.fields, schemaType.isOpen, traceManager, correlationId);
}
exports.typeCheckRecord = typeCheckRecord;
function typeCheckTable(valueType, schemaType, traceManager, correlationId) {
    return typeCheckRecordOrTable(valueType.fields, schemaType.fields, schemaType.isOpen, traceManager, correlationId);
}
exports.typeCheckTable = typeCheckTable;
function typeCheckGenericNumber(valueElements, schemaItemTypes, comparer, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.TypeCheck, typeCheckGenericNumber.name, correlationId);
    const numElements = valueElements.length;
    const numItemTypes = schemaItemTypes.length;
    let upperBound;
    let extraneousIndices;
    let missingIndices;
    if (numElements > numItemTypes) {
        upperBound = numItemTypes;
        extraneousIndices = common_1.ArrayUtils.range(numElements - numItemTypes, numItemTypes);
        missingIndices = [];
    }
    else {
        upperBound = numElements;
        extraneousIndices = [];
        missingIndices = common_1.ArrayUtils.range(numItemTypes - numElements, numElements);
    }
    const validIndices = [];
    const mismatches = new Map();
    for (let index = 0; index < upperBound; index += 1) {
        const element = valueElements[index];
        const schemaItemType = schemaItemTypes[index];
        if (comparer(element, schemaItemType, traceManager, trace.id)) {
            validIndices.push(index);
        }
        else {
            mismatches.set(index, {
                expected: schemaItemType,
                actual: element,
            });
        }
    }
    const result = {
        valid: validIndices,
        invalid: mismatches,
        extraneous: extraneousIndices,
        missing: missingIndices,
    };
    trace.exit();
    return result;
}
function typeCheckRecordOrTable(valueFields, schemaFields, schemaIsOpen, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.TypeCheck, typeCheckRecordOrTable.name, correlationId);
    const validFields = [];
    const mismatches = new Map();
    const extraneousFields = [];
    const missingFields = [...schemaFields.keys()].filter((key) => !valueFields.has(key));
    for (const [key, type] of valueFields.entries()) {
        const schemaValueType = schemaFields.get(key);
        if (schemaValueType !== undefined) {
            if ((0, isCompatible_1.isCompatible)(type, schemaValueType, traceManager, trace.id)) {
                validFields.push(key);
            }
            else {
                mismatches.set(key, {
                    expected: schemaValueType,
                    actual: type,
                });
            }
        }
        else if (schemaIsOpen) {
            validFields.push(key);
        }
        else {
            extraneousFields.push(key);
        }
    }
    const result = {
        valid: validFields,
        invalid: mismatches,
        extraneous: extraneousFields,
        missing: missingFields,
    };
    trace.exit();
    return result;
}
//# sourceMappingURL=typeCheck.js.map