"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.inspectParameter = exports.isValidInvocation = exports.isTypeKind = exports.isTypeInArray = exports.typeKindFromLiteralKind = void 0;
const __1 = require("../..");
const parser_1 = require("../../../parser");
const common_1 = require("../../../common");
const isCompatible_1 = require("./isCompatible");
const isEqualType_1 = require("./isEqualType");
const factories_1 = require("./factories");
const __2 = require("..");
const primitive_1 = require("./primitive");
const typeTraceConstant_1 = require("./typeTraceConstant");
function typeKindFromLiteralKind(literalKind) {
    switch (literalKind) {
        case __1.Ast.LiteralKind.List:
            return __2.Type.TypeKind.List;
        case __1.Ast.LiteralKind.Logical:
            return __2.Type.TypeKind.Logical;
        case __1.Ast.LiteralKind.Null:
            return __2.Type.TypeKind.Null;
        case __1.Ast.LiteralKind.Numeric:
            return __2.Type.TypeKind.Number;
        case __1.Ast.LiteralKind.Record:
            return __2.Type.TypeKind.Record;
        case __1.Ast.LiteralKind.Text:
            return __2.Type.TypeKind.Text;
        default:
            throw common_1.Assert.isNever(literalKind);
    }
}
exports.typeKindFromLiteralKind = typeKindFromLiteralKind;
function isTypeInArray(collection, item) {
    // Fast comparison then deep comparison
    return (collection.includes(item) ||
        collection.find((type) => (0, isEqualType_1.isEqualType)(item, type)) !== undefined);
}
exports.isTypeInArray = isTypeInArray;
function isTypeKind(text) {
    switch (text) {
        case __2.Type.TypeKind.Action:
        case __2.Type.TypeKind.Any:
        case __2.Type.TypeKind.Binary:
        case __2.Type.TypeKind.Date:
        case __2.Type.TypeKind.DateTime:
        case __2.Type.TypeKind.DateTimeZone:
        case __2.Type.TypeKind.Duration:
        case __2.Type.TypeKind.Function:
        case __2.Type.TypeKind.List:
        case __2.Type.TypeKind.Logical:
        case __2.Type.TypeKind.None:
        case __2.Type.TypeKind.NotApplicable:
        case __2.Type.TypeKind.Null:
        case __2.Type.TypeKind.Number:
        case __2.Type.TypeKind.Record:
        case __2.Type.TypeKind.Table:
        case __2.Type.TypeKind.Text:
        case __2.Type.TypeKind.Time:
        case __2.Type.TypeKind.Type:
        case __2.Type.TypeKind.Unknown:
            return true;
        default:
            return false;
    }
}
exports.isTypeKind = isTypeKind;
function isValidInvocation(functionType, args, traceManager, correlationId) {
    const trace = traceManager.entry(typeTraceConstant_1.TypeUtilsTraceConstant.TypeUtils, isValidInvocation.name, correlationId);
    // You can't provide more arguments than are on the function signature.
    if (args.length > functionType.parameters.length) {
        return false;
    }
    const parameters = functionType.parameters;
    const numParameters = parameters.length;
    for (let index = 1; index < numParameters; index += 1) {
        const parameter = common_1.Assert.asDefined(parameters[index]);
        const argType = args[index];
        if (argType !== undefined) {
            const parameterType = (0, factories_1.primitiveType)(parameter.isNullable, common_1.Assert.asDefined(parameter.type));
            if (!(0, isCompatible_1.isCompatible)(argType, parameterType, traceManager, trace.id)) {
                trace.exit();
                return false;
            }
        }
        if (!parameter.isOptional) {
            trace.exit();
            return false;
        }
    }
    trace.exit();
    return true;
}
exports.isValidInvocation = isValidInvocation;
function inspectParameter(nodeIdMapCollection, parameter) {
    switch (parameter.kind) {
        case parser_1.XorNodeKind.Ast:
            return inspectAstParameter(parameter.node);
        case parser_1.XorNodeKind.Context:
            return inspectContextParameter(nodeIdMapCollection, parameter.node);
        default:
            throw common_1.Assert.isNever(parameter);
    }
}
exports.inspectParameter = inspectParameter;
function inspectAstParameter(node) {
    let isNullable;
    let type;
    const parameterType = node.parameterType;
    if (parameterType !== undefined) {
        switch (parameterType.kind) {
            case __1.Ast.NodeKind.AsNullablePrimitiveType: {
                const simplified = __1.AstUtils.simplifyAsNullablePrimitiveType(parameterType);
                isNullable = simplified.isNullable;
                type = (0, primitive_1.typeKindFromPrimitiveTypeConstantKind)(simplified.primitiveTypeConstantKind);
                break;
            }
            case __1.Ast.NodeKind.AsType: {
                const simplified = __1.AstUtils.simplifyType(parameterType.paired);
                isNullable = simplified.isNullable;
                type = (0, primitive_1.typeKindFromPrimitiveTypeConstantKind)(simplified.primitiveTypeConstantKind);
                break;
            }
            default:
                throw common_1.Assert.isNever(parameterType);
        }
    }
    else {
        isNullable = true;
        type = undefined;
    }
    return {
        nameLiteral: node.name.literal,
        isNullable,
        isOptional: node.optionalConstant !== undefined,
        type,
    };
}
function inspectContextParameter(nodeIdMapCollection, parameter) {
    let isNullable;
    let type;
    const name = parser_1.NodeIdMapUtils.nthChildAstChecked(nodeIdMapCollection, parameter.id, 1, __1.Ast.NodeKind.Identifier);
    if (name === undefined) {
        return undefined;
    }
    const optionalConstant = parser_1.NodeIdMapUtils.nthChildAstChecked(nodeIdMapCollection, parameter.id, 0, __1.Ast.NodeKind.Constant);
    const isOptional = optionalConstant !== undefined;
    const parameterType = parser_1.NodeIdMapUtils.nthChildAstChecked(nodeIdMapCollection, parameter.id, 2, __1.Ast.NodeKind.AsNullablePrimitiveType);
    if (parameterType !== undefined) {
        const simplified = __1.AstUtils.simplifyAsNullablePrimitiveType(parameterType);
        isNullable = simplified.isNullable;
        type = (0, primitive_1.typeKindFromPrimitiveTypeConstantKind)(simplified.primitiveTypeConstantKind);
    }
    else {
        isNullable = true;
        type = undefined;
    }
    return {
        nameLiteral: name.literal,
        isOptional,
        isNullable,
        type,
    };
}
//# sourceMappingURL=typeUtils.js.map