"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.errorLineMap = exports.isErrorLine = exports.isErrorState = exports.equalTokens = exports.equalLines = exports.equalStates = exports.tryDeleteLine = exports.tryUpdateRange = exports.tryUpdateLine = exports.tryAppendLine = exports.tryLex = exports.LineMode = exports.LineKind = void 0;
const common_1 = require("../common");
const language_1 = require("../language");
const _1 = require(".");
var LineKind;
(function (LineKind) {
    LineKind["Error"] = "Error";
    LineKind["Touched"] = "Touched";
    LineKind["TouchedWithError"] = "TouchedWithError";
    LineKind["Untouched"] = "Untouched";
})(LineKind = exports.LineKind || (exports.LineKind = {}));
// A line's LineMode determines how its tokenized.
// There are two broad categories for a LineMode:
//  * Atomic (single) line tokenization (default)
//  * Continuation of a multiline token (eg. multiline comment)
//
// All LineModes other than Default are a type of multiline tokenization.
var LineMode;
(function (LineMode) {
    LineMode["Comment"] = "Comment";
    LineMode["Default"] = "Default";
    LineMode["QuotedIdentifier"] = "QuotedIdentifier";
    LineMode["Text"] = "Text";
})(LineMode = exports.LineMode || (exports.LineMode = {}));
// This export is a Result ensuring wrapper around the un-exported implementation.
function tryLex(settings, text) {
    return ensureCommonOrLexerResult(() => lex(settings, text), settings.locale);
}
exports.tryLex = tryLex;
// This export is a Result ensuring wrapper around the un-exported implementation.
function tryAppendLine(state, text, lineTerminator) {
    return ensureCommonOrLexerResult(() => appendLine(state, text, lineTerminator), state.locale);
}
exports.tryAppendLine = tryAppendLine;
// This export is a Result ensuring wrapper around the un-exported implementation.
function tryUpdateLine(state, lineNumber, text) {
    return ensureCommonOrLexerResult(() => updateLine(state, lineNumber, text), state.locale);
}
exports.tryUpdateLine = tryUpdateLine;
// This export is a Result ensuring wrapper around the un-exported implementation.
function tryUpdateRange(state, range, text) {
    return ensureCommonOrLexerResult(() => updateRange(state, range, text), state.locale);
}
exports.tryUpdateRange = tryUpdateRange;
// This export is a Result ensuring wrapper around the un-exported implementation.
function tryDeleteLine(state, lineNumber) {
    return ensureCommonOrLexerResult(() => deleteLine(state, lineNumber), state.locale);
}
exports.tryDeleteLine = tryDeleteLine;
// deep state comparison
function equalStates(leftState, rightState) {
    return equalLines(leftState.lines, rightState.lines);
}
exports.equalStates = equalStates;
// deep line comparison
// partial equality as ILine.text is ignored
function equalLines(leftLines, rightLines) {
    if (leftLines.length !== rightLines.length) {
        return false;
    }
    const numLines = leftLines.length;
    for (let lineIndex = 0; lineIndex < numLines; lineIndex += 1) {
        const left = leftLines[lineIndex];
        const right = rightLines[lineIndex];
        const leftTokens = left.tokens;
        const rightTokens = right.tokens;
        const isEqualQuickCheck = left.kind === right.kind &&
            left.lineTerminator === right.lineTerminator &&
            left.lineModeStart === right.lineModeStart &&
            left.lineModeEnd === right.lineModeEnd &&
            leftTokens.length === rightTokens.length;
        if (!isEqualQuickCheck) {
            return false;
        }
        // isEqualQuickCheck ensures tokens.length is the same
        const numTokens = leftTokens.length;
        for (let tokenIndex = 0; tokenIndex < numTokens; tokenIndex += 1) {
            if (!equalTokens(leftTokens[tokenIndex], rightTokens[tokenIndex])) {
                return false;
            }
        }
    }
    return true;
}
exports.equalLines = equalLines;
// deep token comparison
function equalTokens(leftToken, rightToken) {
    return (leftToken.kind === rightToken.kind &&
        leftToken.data === rightToken.data &&
        leftToken.positionStart === rightToken.positionStart &&
        leftToken.positionEnd === rightToken.positionEnd);
}
exports.equalTokens = equalTokens;
function isErrorState(state) {
    const linesWithErrors = state.lines.filter(isErrorLine);
    return linesWithErrors.length !== 0;
}
exports.isErrorState = isErrorState;
function isErrorLine(line) {
    switch (line.kind) {
        case LineKind.Error:
        case LineKind.TouchedWithError:
            return true;
        case LineKind.Touched:
        case LineKind.Untouched:
            return false;
        default:
            throw common_1.Assert.isNever(line);
    }
}
exports.isErrorLine = isErrorLine;
function errorLineMap(state) {
    const errorLines = state.lines.reduce((errorLineMap, line, index) => {
        if (isErrorLine(line)) {
            errorLineMap.set(index, line);
        }
        return errorLineMap;
    }, new Map());
    return errorLines.size !== 0 ? errorLines : undefined;
}
exports.errorLineMap = errorLineMap;
// Takes a string and splits it on all valid Power Query terminators.
// The split lines retain what newline was used to create the split.
function splitOnLineTerminators(startingText) {
    let lines = startingText.split("\r\n").map((lineText) => ({
        text: lineText,
        lineTerminator: "\r\n",
    }));
    const lineTerminators = [
        "\n",
        "\u2028",
        "\u2029", // PARAGRAPH SEPARATOR
    ];
    let index = 0;
    while (index < lines.length) {
        let indexWasExpanded = false;
        for (const lineTerminator of lineTerminators) {
            const splitLine = lines[index];
            const text = splitLine.text;
            if (text.indexOf(lineTerminator) !== -1) {
                indexWasExpanded = true;
                const split = text.split(lineTerminator).map((lineText) => ({
                    text: lineText,
                    lineTerminator,
                }));
                split[split.length - 1].lineTerminator = splitLine.lineTerminator;
                lines = [...lines.slice(0, index), ...split, ...lines.slice(index + 1)];
            }
        }
        if (!indexWasExpanded) {
            index += 1;
        }
    }
    lines[lines.length - 1].lineTerminator = "";
    return lines;
}
function ensureCommonOrLexerResult(functionToWrap, locale) {
    try {
        return common_1.ResultUtils.ok(functionToWrap());
    }
    catch (caught) {
        common_1.Assert.isInstanceofError(caught);
        let convertedError;
        if (_1.LexError.isTInnerLexError(caught)) {
            convertedError = new _1.LexError.LexError(caught);
        }
        else {
            convertedError = common_1.CommonError.ensureCommonError(caught, locale);
        }
        return common_1.ResultUtils.error(convertedError);
    }
}
function lex(settings, text) {
    const splitLines = splitOnLineTerminators(text);
    const tokenizedLines = tokenizedLinesFrom(splitLines, LineMode.Default, settings.locale, settings.cancellationToken);
    return {
        lines: tokenizedLines,
        locale: settings.locale,
        cancellationToken: settings.cancellationToken,
    };
}
function appendLine(state, text, lineTerminator) {
    var _a;
    (_a = state.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
    const lines = state.lines;
    const numLines = lines.length;
    const latestLine = lines[numLines - 1];
    const lineModeStart = latestLine ? latestLine.lineModeEnd : LineMode.Default;
    const untokenizedLine = lineFrom(text, lineTerminator, lineModeStart);
    const tokenizedLine = tokenize(untokenizedLine, numLines, state.locale, state.cancellationToken);
    return Object.assign(Object.assign({}, state), { lines: state.lines.concat(tokenizedLine) });
}
function updateLine(state, lineNumber, text) {
    var _a;
    (_a = state.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
    const error = testLineNumberError(state, lineNumber);
    if (error !== undefined) {
        throw error;
    }
    const line = state.lines[lineNumber];
    const range = rangeFrom(line, lineNumber);
    return updateRange(state, range, text);
}
function updateRange(state, range, text) {
    var _a, _b;
    (_a = state.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
    const error = testBadRangeError(state, range);
    if (error !== undefined) {
        throw error;
    }
    const splitLines = splitOnLineTerminators(text);
    const rangeStart = range.start;
    const lineStart = state.lines[rangeStart.lineNumber];
    const textPrefix = lineStart.text.substring(0, rangeStart.lineCodeUnit);
    splitLines[0].text = textPrefix + splitLines[0].text;
    const rangeEnd = range.end;
    const lineEnd = state.lines[rangeEnd.lineNumber];
    const textSuffix = lineEnd.text.substr(rangeEnd.lineCodeUnit);
    const lastSplitLine = splitLines[splitLines.length - 1];
    lastSplitLine.text = lastSplitLine.text + textSuffix;
    // make sure we have a line terminator
    lastSplitLine.lineTerminator = lineEnd.lineTerminator;
    const previousLine = state.lines[rangeStart.lineNumber - 1];
    const previousLineModeEnd = (_b = previousLine === null || previousLine === void 0 ? void 0 : previousLine.lineModeEnd) !== null && _b !== void 0 ? _b : LineMode.Default;
    const newLines = tokenizedLinesFrom(splitLines, previousLineModeEnd, state.locale, state.cancellationToken);
    const lines = [
        ...state.lines.slice(0, rangeStart.lineNumber),
        ...newLines,
        ...retokenizeLines(state, rangeEnd.lineNumber + 1, newLines[newLines.length - 1].lineModeEnd),
    ];
    return {
        lines,
        locale: state.locale,
        cancellationToken: state.cancellationToken,
    };
}
function deleteLine(state, lineNumber) {
    var _a;
    (_a = state.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
    const error = testLineNumberError(state, lineNumber);
    if (error !== undefined) {
        throw error;
    }
    return Object.assign(Object.assign({}, state), { lines: common_1.ArrayUtils.assertRemoveAtIndex(state.lines, lineNumber) });
}
function lineFrom(text, lineTerminator, lineModeStart) {
    return {
        kind: LineKind.Untouched,
        text,
        lineTerminator,
        lineModeStart,
        lineModeEnd: LineMode.Default,
        tokens: [],
    };
}
function graphemePositionFrom(text, lineNumber, lineCodeUnit) {
    return common_1.StringUtils.graphemePositionFrom(text, lineCodeUnit, lineNumber, undefined);
}
function rangeFrom(line, lineNumber) {
    return {
        start: {
            lineNumber,
            lineCodeUnit: 0,
        },
        end: {
            lineNumber,
            lineCodeUnit: line.text.length,
        },
    };
}
function tokenizedLinesFrom(splitLines, previousLineModeEnd, locale, cancellationToken) {
    const tokenizedLines = [];
    for (const [splitLine, index] of common_1.ArrayUtils.enumerate(splitLines)) {
        const untokenizedLine = lineFrom(splitLine.text, splitLine.lineTerminator, previousLineModeEnd);
        const tokenizedLine = tokenize(untokenizedLine, index, locale, cancellationToken);
        tokenizedLines.push(tokenizedLine);
        previousLineModeEnd = tokenizedLine.lineModeEnd;
    }
    return tokenizedLines;
}
// If an earlier line changed its lineModeEnd, eg. inserting a `"` to start a string literal,
// then the proceeding lines would need to be retokenized.
// Stops retokenizing when previous.lineModeEnd !== current.lineModeStart.
// Returns lines in the range [lineNumber, lines.length -1]
function retokenizeLines(state, lineNumber, previousLineModeEnd) {
    const lines = state.lines;
    if (lines[lineNumber] === undefined) {
        return [];
    }
    const retokenizedLines = [];
    if (previousLineModeEnd !== lines[lineNumber].lineModeStart) {
        const offsetLineNumber = lineNumber;
        let currentLine = lines[lineNumber];
        while (currentLine) {
            const line = currentLine;
            if (previousLineModeEnd !== line.lineModeStart) {
                const untokenizedLine = lineFrom(line.text, line.lineTerminator, previousLineModeEnd);
                const retokenizedLine = tokenize(untokenizedLine, offsetLineNumber, state.locale, state.cancellationToken);
                retokenizedLines.push(retokenizedLine);
                previousLineModeEnd = retokenizedLine.lineModeEnd;
                lineNumber += 1;
                currentLine = lines[lineNumber];
            }
            else {
                return [...retokenizedLines, ...lines.slice(lineNumber + 1)];
            }
        }
        return retokenizedLines;
    }
    else {
        return lines.slice(lineNumber);
    }
}
// The main function of the lexer's tokenizer.
function tokenize(line, lineNumber, locale, cancellationToken) {
    cancellationToken === null || cancellationToken === void 0 ? void 0 : cancellationToken.throwIfCancelled();
    switch (line.kind) {
        // Cannot tokenize something that ended with an error,
        // nothing has changed since the last tokenize.
        // Update the line's text before trying again.
        case LineKind.Error:
            return line;
        case LineKind.Touched:
            // The line was already fully lexed once.
            // Without any text changes it should throw eof to help diagnose
            // why it's trying to retokenize.
            return Object.assign(Object.assign({}, line), { kind: LineKind.Error, error: new _1.LexError.LexError(new _1.LexError.EndOfStreamError(locale)) });
        // Cannot tokenize something that previously ended with an error.
        // Update the line's text before trying again.
        case LineKind.TouchedWithError:
            return {
                kind: LineKind.Error,
                text: line.text,
                lineTerminator: line.lineTerminator,
                lineModeStart: line.lineModeStart,
                lineModeEnd: line.lineModeEnd,
                tokens: line.tokens,
                error: new _1.LexError.LexError(new _1.LexError.BadStateError(line.error, locale)),
            };
        case LineKind.Untouched:
            break;
        default:
            throw common_1.Assert.isNever(line);
    }
    const untouchedLine = line;
    const text = untouchedLine.text;
    const textLength = text.length;
    // If there's nothing to tokenize set lineModeEnd to lineModeStart.
    if (textLength === 0) {
        return {
            kind: LineKind.Touched,
            text: line.text,
            lineTerminator: line.lineTerminator,
            lineModeStart: line.lineModeStart,
            lineModeEnd: line.lineModeStart,
            tokens: [],
        };
    }
    let lineMode = line.lineModeStart;
    let currentPosition = 0;
    if (lineMode === LineMode.Default) {
        currentPosition = drainWhitespace(text, currentPosition);
    }
    const newTokens = [];
    let continueLexing = currentPosition !== text.length;
    let lexError;
    // While neither eof nor having encountered an error:
    //  * Lex according to lineModeStart, starting from currentPosition.
    //  * Update currentPosition and lineMode.
    //  * Drain whitespace.
    while (continueLexing) {
        cancellationToken === null || cancellationToken === void 0 ? void 0 : cancellationToken.throwIfCancelled();
        try {
            let readOutcome;
            switch (lineMode) {
                case LineMode.Comment:
                    readOutcome = tokenizeMultilineCommentContentOrEnd(line, currentPosition);
                    break;
                case LineMode.Default:
                    readOutcome = tokenizeDefault(line, lineNumber, currentPosition, locale);
                    break;
                case LineMode.QuotedIdentifier:
                    readOutcome = tokenizeQuotedIdentifierContentOrEnd(line, currentPosition);
                    break;
                case LineMode.Text:
                    readOutcome = tokenizeTextLiteralContentOrEnd(line, currentPosition);
                    break;
                default:
                    throw common_1.Assert.isNever(lineMode);
            }
            lineMode = readOutcome.lineMode;
            const token = readOutcome.token;
            newTokens.push(token);
            if (lineMode === LineMode.Default) {
                currentPosition = drainWhitespace(text, token.positionEnd);
            }
            else {
                currentPosition = token.positionEnd;
            }
            if (currentPosition === textLength) {
                continueLexing = false;
            }
        }
        catch (caught) {
            let error;
            if (_1.LexError.isTInnerLexError(caught)) {
                error = new _1.LexError.LexError(caught);
            }
            else {
                common_1.Assert.isInstanceofError(caught);
                error = common_1.CommonError.ensureCommonError(caught, locale);
            }
            continueLexing = false;
            lexError = error;
        }
    }
    let partialTokenizeResult;
    if (lexError) {
        if (newTokens.length) {
            partialTokenizeResult = common_1.PartialResultUtils.incomplete({
                tokenizeChanges: {
                    tokens: newTokens,
                    lineModeEnd: lineMode,
                },
                error: lexError,
            });
        }
        else {
            partialTokenizeResult = common_1.PartialResultUtils.error(lexError);
        }
    }
    else {
        partialTokenizeResult = common_1.PartialResultUtils.ok({
            tokens: newTokens,
            lineModeEnd: lineMode,
        });
    }
    return updateLineState(line, partialTokenizeResult);
}
// Takes the return from a tokenizeX function to updates the TLine's state.
function updateLineState(line, potentiallyIncompleteResult) {
    switch (potentiallyIncompleteResult.kind) {
        case common_1.PartialResultKind.Ok: {
            const tokenizeChanges = potentiallyIncompleteResult.value;
            const newTokens = line.tokens.concat(tokenizeChanges.tokens);
            return {
                kind: LineKind.Touched,
                text: line.text,
                lineTerminator: line.lineTerminator,
                lineModeStart: line.lineModeStart,
                lineModeEnd: tokenizeChanges.lineModeEnd,
                tokens: newTokens,
            };
        }
        case common_1.PartialResultKind.Incomplete: {
            const tokenizeChanges = potentiallyIncompleteResult.partial.tokenizeChanges;
            const newTokens = line.tokens.concat(tokenizeChanges.tokens);
            return {
                kind: LineKind.TouchedWithError,
                text: line.text,
                lineTerminator: line.lineTerminator,
                lineModeStart: line.lineModeStart,
                lineModeEnd: tokenizeChanges.lineModeEnd,
                tokens: newTokens,
                error: potentiallyIncompleteResult.partial.error,
            };
        }
        case common_1.PartialResultKind.Error:
            return {
                kind: LineKind.Error,
                text: line.text,
                lineModeStart: line.lineModeStart,
                lineTerminator: line.lineTerminator,
                lineModeEnd: line.lineModeEnd,
                tokens: line.tokens,
                error: potentiallyIncompleteResult.error,
            };
        default:
            throw common_1.Assert.isNever(potentiallyIncompleteResult);
    }
}
// Read until either "*/" or eof
function tokenizeMultilineCommentContentOrEnd(line, positionStart) {
    const text = line.text;
    const indexOfCloseComment = text.indexOf("*/", positionStart);
    if (indexOfCloseComment === -1) {
        return {
            token: readRestOfLine(language_1.Token.LineTokenKind.MultilineCommentContent, text, positionStart),
            lineMode: LineMode.Comment,
        };
    }
    else {
        const positionEnd = indexOfCloseComment + 2;
        return {
            token: readTokenFrom(language_1.Token.LineTokenKind.MultilineCommentEnd, text, positionStart, positionEnd),
            lineMode: LineMode.Default,
        };
    }
}
// Read until either string literal end or eof
function tokenizeQuotedIdentifierContentOrEnd(line, currentPosition) {
    const read = tokenizeTextLiteralContentOrEnd(line, currentPosition);
    switch (read.token.kind) {
        case language_1.Token.LineTokenKind.TextLiteralContent:
            return {
                lineMode: LineMode.QuotedIdentifier,
                token: Object.assign(Object.assign({}, read.token), { kind: language_1.Token.LineTokenKind.QuotedIdentifierContent }),
            };
        case language_1.Token.LineTokenKind.TextLiteralEnd:
            return {
                lineMode: LineMode.Default,
                token: Object.assign(Object.assign({}, read.token), { kind: language_1.Token.LineTokenKind.QuotedIdentifierEnd }),
            };
        default:
            throw new common_1.CommonError.InvariantError(`expected the return to be either ${language_1.Token.LineTokenKind.TextLiteralContent} or ${language_1.Token.LineTokenKind.TextLiteralEnd}`, { read });
    }
}
// Read until either string literal end or eof
function tokenizeTextLiteralContentOrEnd(line, currentPosition) {
    const text = line.text;
    const positionEnd = indexOfTextEnd(text, currentPosition);
    if (positionEnd === undefined) {
        return {
            token: readRestOfLine(language_1.Token.LineTokenKind.TextLiteralContent, text, currentPosition),
            lineMode: LineMode.Text,
        };
    }
    else {
        return {
            token: readTokenFrom(language_1.Token.LineTokenKind.TextLiteralEnd, text, currentPosition, positionEnd + 1),
            lineMode: LineMode.Default,
        };
    }
}
function tokenizeDefault(line, lineNumber, positionStart, locale) {
    const text = line.text;
    const chr1 = text[positionStart];
    let token;
    let lineMode = LineMode.Default;
    if (chr1 === "!") {
        token = readConstant(language_1.Token.LineTokenKind.Bang, text, positionStart, 1);
    }
    else if (chr1 === "&") {
        token = readConstant(language_1.Token.LineTokenKind.Ampersand, text, positionStart, 1);
    }
    else if (chr1 === "(") {
        token = readConstant(language_1.Token.LineTokenKind.LeftParenthesis, text, positionStart, 1);
    }
    else if (chr1 === ")") {
        token = readConstant(language_1.Token.LineTokenKind.RightParenthesis, text, positionStart, 1);
    }
    else if (chr1 === "*") {
        token = readConstant(language_1.Token.LineTokenKind.Asterisk, text, positionStart, 1);
    }
    else if (chr1 === "+") {
        token = readConstant(language_1.Token.LineTokenKind.Plus, text, positionStart, 1);
    }
    else if (chr1 === ",") {
        token = readConstant(language_1.Token.LineTokenKind.Comma, text, positionStart, 1);
    }
    else if (chr1 === "-") {
        token = readConstant(language_1.Token.LineTokenKind.Minus, text, positionStart, 1);
    }
    else if (chr1 === ";") {
        token = readConstant(language_1.Token.LineTokenKind.Semicolon, text, positionStart, 1);
    }
    else if (chr1 === "?") {
        const chr2 = text[positionStart + 1];
        if (chr2 === "?") {
            token = readConstant(language_1.Token.LineTokenKind.NullCoalescingOperator, text, positionStart, 2);
        }
        else {
            token = readConstant(language_1.Token.LineTokenKind.QuestionMark, text, positionStart, 1);
        }
    }
    else if (chr1 === "@") {
        token = readConstant(language_1.Token.LineTokenKind.AtSign, text, positionStart, 1);
    }
    else if (chr1 === "[") {
        token = readConstant(language_1.Token.LineTokenKind.LeftBracket, text, positionStart, 1);
    }
    else if (chr1 === "]") {
        token = readConstant(language_1.Token.LineTokenKind.RightBracket, text, positionStart, 1);
    }
    else if (chr1 === "{") {
        token = readConstant(language_1.Token.LineTokenKind.LeftBrace, text, positionStart, 1);
    }
    else if (chr1 === "}") {
        token = readConstant(language_1.Token.LineTokenKind.RightBrace, text, positionStart, 1);
    }
    else if (chr1 === '"') {
        const read = readOrStartTextLiteral(text, positionStart);
        token = read.token;
        lineMode = read.lineMode;
    }
    else if (chr1 === "0") {
        const chr2 = text[positionStart + 1];
        if (chr2 === "x" || chr2 === "X") {
            token = readHexLiteral(text, lineNumber, positionStart, locale);
        }
        else {
            token = readNumericLiteral(text, lineNumber, positionStart, locale);
        }
    }
    else if ("1" <= chr1 && chr1 <= "9") {
        token = readNumericLiteral(text, lineNumber, positionStart, locale);
    }
    else if (chr1 === ".") {
        const chr2 = text[positionStart + 1];
        if (chr2 === undefined) {
            throw new _1.LexError.UnexpectedEofError(graphemePositionFrom(text, lineNumber, positionStart), locale);
        }
        else if ("1" <= chr2 && chr2 <= "9") {
            token = readNumericLiteral(text, lineNumber, positionStart, locale);
        }
        else if (chr2 === ".") {
            const chr3 = text[positionStart + 2];
            if (chr3 === ".") {
                token = readConstant(language_1.Token.LineTokenKind.Ellipsis, text, positionStart, 3);
            }
            else {
                token = readConstant(language_1.Token.LineTokenKind.DotDot, text, positionStart, 2);
            }
        }
        else {
            throw unexpectedReadError(locale, text, lineNumber, positionStart);
        }
    }
    else if (chr1 === ">") {
        const chr2 = text[positionStart + 1];
        if (chr2 === "=") {
            token = readConstant(language_1.Token.LineTokenKind.GreaterThanEqualTo, text, positionStart, 2);
        }
        else {
            token = readConstant(language_1.Token.LineTokenKind.GreaterThan, text, positionStart, 1);
        }
    }
    else if (chr1 === "<") {
        const chr2 = text[positionStart + 1];
        if (chr2 === "=") {
            token = readConstant(language_1.Token.LineTokenKind.LessThanEqualTo, text, positionStart, 2);
        }
        else if (chr2 === ">") {
            token = readConstant(language_1.Token.LineTokenKind.NotEqual, text, positionStart, 2);
        }
        else {
            token = readConstant(language_1.Token.LineTokenKind.LessThan, text, positionStart, 1);
        }
    }
    else if (chr1 === "=") {
        const chr2 = text[positionStart + 1];
        if (chr2 === ">") {
            token = readConstant(language_1.Token.LineTokenKind.FatArrow, text, positionStart, 2);
        }
        else {
            token = readConstant(language_1.Token.LineTokenKind.Equal, text, positionStart, 1);
        }
    }
    else if (chr1 === "/") {
        const chr2 = text[positionStart + 1];
        if (chr2 === "/") {
            token = readLineComment(text, positionStart);
        }
        else if (chr2 === "*") {
            const read = readOrStartMultilineComment(text, positionStart);
            token = read.token;
            lineMode = read.lineMode;
        }
        else {
            token = readConstant(language_1.Token.LineTokenKind.Division, text, positionStart, 1);
        }
    }
    else if (chr1 === "#") {
        const chr2 = text[positionStart + 1];
        if (chr2 === '"') {
            const read = readOrStartQuotedIdentifier(text, positionStart);
            token = read.token;
            lineMode = read.lineMode;
        }
        else {
            token = readKeyword(text, lineNumber, positionStart, locale);
        }
    }
    else {
        token = readKeywordOrIdentifier(text, lineNumber, positionStart, locale);
    }
    return {
        token,
        lineMode,
    };
}
// newlines are not considered whitespace
function drainWhitespace(text, position) {
    let continueDraining = text[position] !== undefined;
    while (continueDraining) {
        const length = common_1.StringUtils.regexMatchLength(common_1.Pattern.Whitespace, text, position);
        if (length) {
            position += length;
        }
        else {
            continueDraining = false;
        }
    }
    return position;
}
function readOrStartTextLiteral(text, currentPosition) {
    const positionEnd = indexOfTextEnd(text, currentPosition + 1);
    if (positionEnd !== undefined) {
        return {
            token: readTokenFrom(language_1.Token.LineTokenKind.TextLiteral, text, currentPosition, positionEnd + 1),
            lineMode: LineMode.Default,
        };
    }
    else {
        return {
            token: readRestOfLine(language_1.Token.LineTokenKind.TextLiteralStart, text, currentPosition),
            lineMode: LineMode.Text,
        };
    }
}
function readHexLiteral(text, lineNumber, positionStart, locale) {
    const positionEnd = indexOfRegexEnd(common_1.Pattern.Hex, text, positionStart);
    if (positionEnd === undefined) {
        throw new _1.LexError.ExpectedError(graphemePositionFrom(text, lineNumber, positionStart), _1.LexError.ExpectedKind.HexLiteral, locale);
    }
    return readTokenFrom(language_1.Token.LineTokenKind.HexLiteral, text, positionStart, positionEnd);
}
function readNumericLiteral(text, lineNumber, positionStart, locale) {
    const positionEnd = indexOfRegexEnd(common_1.Pattern.Numeric, text, positionStart);
    if (positionEnd === undefined) {
        throw new _1.LexError.ExpectedError(graphemePositionFrom(text, lineNumber, positionStart), _1.LexError.ExpectedKind.Numeric, locale);
    }
    return readTokenFrom(language_1.Token.LineTokenKind.NumericLiteral, text, positionStart, positionEnd);
}
function readLineComment(text, positionStart) {
    return readRestOfLine(language_1.Token.LineTokenKind.LineComment, text, positionStart);
}
function readOrStartMultilineComment(text, positionStart) {
    const indexOfCloseComment = text.indexOf("*/", positionStart + 2);
    if (indexOfCloseComment === -1) {
        return {
            token: readRestOfLine(language_1.Token.LineTokenKind.MultilineCommentStart, text, positionStart),
            lineMode: LineMode.Comment,
        };
    }
    else {
        const positionEnd = indexOfCloseComment + 2;
        return {
            token: readTokenFrom(language_1.Token.LineTokenKind.MultilineComment, text, positionStart, positionEnd),
            lineMode: LineMode.Default,
        };
    }
}
function readKeyword(text, lineNumber, positionStart, locale) {
    const lineToken = readKeywordHelper(text, positionStart);
    if (lineToken) {
        return lineToken;
    }
    else {
        throw unexpectedReadError(locale, text, lineNumber, positionStart);
    }
}
function readKeywordHelper(text, currentPosition) {
    const identifierPositionStart = text[currentPosition] === "#" ? currentPosition + 1 : currentPosition;
    const identifierPositionEnd = indexOfIdentifierEnd(text, identifierPositionStart);
    if (identifierPositionEnd === undefined) {
        return undefined;
    }
    const data = text.substring(currentPosition, identifierPositionEnd);
    const keywordTokenKind = keywordLineTokenKindFrom(data);
    if (keywordTokenKind === undefined) {
        return undefined;
    }
    else {
        return {
            kind: keywordTokenKind,
            positionStart: currentPosition,
            positionEnd: identifierPositionEnd,
            data,
        };
    }
}
function readOrStartQuotedIdentifier(text, currentPosition) {
    const positionEnd = indexOfTextEnd(text, currentPosition + 2);
    if (positionEnd !== undefined) {
        return {
            token: readTokenFrom(language_1.Token.LineTokenKind.Identifier, text, currentPosition, positionEnd + 1),
            lineMode: LineMode.Default,
        };
    }
    else {
        return {
            token: readRestOfLine(language_1.Token.LineTokenKind.QuotedIdentifierStart, text, currentPosition),
            lineMode: LineMode.QuotedIdentifier,
        };
    }
}
// The case for quoted identifier has already been taken care of.
// The null-literal is also read here.
function readKeywordOrIdentifier(text, lineNumber, positionStart, locale) {
    // keyword
    if (text[positionStart] === "#") {
        return readKeyword(text, lineNumber, positionStart, locale);
    }
    // either keyword or identifier
    else {
        const positionEnd = indexOfIdentifierEnd(text, positionStart);
        if (positionEnd === undefined) {
            throw new _1.LexError.ExpectedError(graphemePositionFrom(text, lineNumber, positionStart), _1.LexError.ExpectedKind.KeywordOrIdentifier, locale);
        }
        const data = text.substring(positionStart, positionEnd);
        const keywordTokenKind = keywordLineTokenKindFrom(data);
        let tokenKind;
        if (keywordTokenKind !== undefined) {
            tokenKind = keywordTokenKind;
        }
        else if (data === "null") {
            tokenKind = language_1.Token.LineTokenKind.NullLiteral;
        }
        else {
            tokenKind = language_1.Token.LineTokenKind.Identifier;
        }
        return {
            kind: tokenKind,
            positionStart,
            positionEnd,
            data,
        };
    }
}
function readConstant(lineTokenKind, text, positionStart, length) {
    const positionEnd = positionStart + length;
    return readTokenFrom(lineTokenKind, text, positionStart, positionEnd);
}
function readTokenFrom(lineTokenKind, text, positionStart, positionEnd) {
    return {
        kind: lineTokenKind,
        positionStart,
        positionEnd,
        data: text.substring(positionStart, positionEnd),
    };
}
function readRestOfLine(lineTokenKind, text, positionStart) {
    const positionEnd = text.length;
    return readTokenFrom(lineTokenKind, text, positionStart, positionEnd);
}
function indexOfRegexEnd(pattern, text, positionStart) {
    const length = common_1.StringUtils.regexMatchLength(pattern, text, positionStart);
    return length !== undefined ? positionStart + length : undefined;
}
function indexOfIdentifierEnd(text, positionStart) {
    const length = language_1.IdentifierUtils.getIdentifierLength(text, positionStart, true);
    return length !== undefined ? positionStart + length : undefined;
}
function keywordLineTokenKindFrom(data) {
    switch (data) {
        case language_1.Keyword.KeywordKind.And:
            return language_1.Token.LineTokenKind.KeywordAnd;
        case language_1.Keyword.KeywordKind.As:
            return language_1.Token.LineTokenKind.KeywordAs;
        case language_1.Keyword.KeywordKind.Each:
            return language_1.Token.LineTokenKind.KeywordEach;
        case language_1.Keyword.KeywordKind.Else:
            return language_1.Token.LineTokenKind.KeywordElse;
        case language_1.Keyword.KeywordKind.Error:
            return language_1.Token.LineTokenKind.KeywordError;
        case language_1.Keyword.KeywordKind.False:
            return language_1.Token.LineTokenKind.KeywordFalse;
        case language_1.Keyword.KeywordKind.If:
            return language_1.Token.LineTokenKind.KeywordIf;
        case language_1.Keyword.KeywordKind.In:
            return language_1.Token.LineTokenKind.KeywordIn;
        case language_1.Keyword.KeywordKind.Is:
            return language_1.Token.LineTokenKind.KeywordIs;
        case language_1.Keyword.KeywordKind.Let:
            return language_1.Token.LineTokenKind.KeywordLet;
        case language_1.Keyword.KeywordKind.Meta:
            return language_1.Token.LineTokenKind.KeywordMeta;
        case language_1.Keyword.KeywordKind.Not:
            return language_1.Token.LineTokenKind.KeywordNot;
        case language_1.Keyword.KeywordKind.Or:
            return language_1.Token.LineTokenKind.KeywordOr;
        case language_1.Keyword.KeywordKind.Otherwise:
            return language_1.Token.LineTokenKind.KeywordOtherwise;
        case language_1.Keyword.KeywordKind.Section:
            return language_1.Token.LineTokenKind.KeywordSection;
        case language_1.Keyword.KeywordKind.Shared:
            return language_1.Token.LineTokenKind.KeywordShared;
        case language_1.Keyword.KeywordKind.Then:
            return language_1.Token.LineTokenKind.KeywordThen;
        case language_1.Keyword.KeywordKind.True:
            return language_1.Token.LineTokenKind.KeywordTrue;
        case language_1.Keyword.KeywordKind.Try:
            return language_1.Token.LineTokenKind.KeywordTry;
        case language_1.Keyword.KeywordKind.Type:
            return language_1.Token.LineTokenKind.KeywordType;
        case language_1.Keyword.KeywordKind.HashBinary:
            return language_1.Token.LineTokenKind.KeywordHashBinary;
        case language_1.Keyword.KeywordKind.HashDate:
            return language_1.Token.LineTokenKind.KeywordHashDate;
        case language_1.Keyword.KeywordKind.HashDateTime:
            return language_1.Token.LineTokenKind.KeywordHashDateTime;
        case language_1.Keyword.KeywordKind.HashDateTimeZone:
            return language_1.Token.LineTokenKind.KeywordHashDateTimeZone;
        case language_1.Keyword.KeywordKind.HashDuration:
            return language_1.Token.LineTokenKind.KeywordHashDuration;
        case language_1.Keyword.KeywordKind.HashInfinity:
            return language_1.Token.LineTokenKind.KeywordHashInfinity;
        case language_1.Keyword.KeywordKind.HashNan:
            return language_1.Token.LineTokenKind.KeywordHashNan;
        case language_1.Keyword.KeywordKind.HashSections:
            return language_1.Token.LineTokenKind.KeywordHashSections;
        case language_1.Keyword.KeywordKind.HashShared:
            return language_1.Token.LineTokenKind.KeywordHashShared;
        case language_1.Keyword.KeywordKind.HashTable:
            return language_1.Token.LineTokenKind.KeywordHashTable;
        case language_1.Keyword.KeywordKind.HashTime:
            return language_1.Token.LineTokenKind.KeywordHashTime;
        default:
            return undefined;
    }
}
function indexOfTextEnd(text, positionStart) {
    let indexLow = positionStart;
    let positionEnd = text.indexOf('"', indexLow);
    while (positionEnd !== -1) {
        if (text[positionEnd + 1] === '"') {
            indexLow = positionEnd + 2;
            positionEnd = text.indexOf('"', indexLow);
        }
        else {
            return positionEnd;
        }
    }
    return undefined;
}
function unexpectedReadError(locale, text, lineNumber, lineCodeUnit) {
    return new _1.LexError.UnexpectedReadError(graphemePositionFrom(text, lineNumber, lineCodeUnit), locale);
}
function testLineNumberError(state, lineNumber) {
    const numLines = state.lines.length;
    if (lineNumber >= numLines) {
        return new _1.LexError.BadLineNumberError(_1.LexError.BadLineNumberKind.GreaterThanNumLines, lineNumber, numLines, state.locale);
    }
    else if (lineNumber < 0) {
        return new _1.LexError.BadLineNumberError(_1.LexError.BadLineNumberKind.LessThanZero, lineNumber, numLines, state.locale);
    }
    else {
        return undefined;
    }
}
// Validator for Range.
function testBadRangeError(state, range) {
    const start = range.start;
    const end = range.end;
    const numLines = state.lines.length;
    let kind;
    if (start.lineNumber === end.lineNumber && start.lineCodeUnit > end.lineCodeUnit) {
        kind = _1.LexError.BadRangeKind.SameLine_LineCodeUnitStart_Higher;
    }
    else if (start.lineNumber > end.lineNumber) {
        kind = _1.LexError.BadRangeKind.LineNumberStart_GreaterThan_LineNumberEnd;
    }
    else if (start.lineNumber < 0) {
        kind = _1.LexError.BadRangeKind.LineNumberStart_LessThan_Zero;
    }
    else if (start.lineNumber >= numLines) {
        kind = _1.LexError.BadRangeKind.LineNumberStart_GreaterThan_NumLines;
    }
    else if (end.lineNumber >= numLines) {
        kind = _1.LexError.BadRangeKind.LineNumberEnd_GreaterThan_NumLines;
    }
    if (kind !== undefined) {
        return new _1.LexError.BadRangeError(range, kind, state.locale);
    }
    const lines = state.lines;
    const rangeStart = range.start;
    const rangeEnd = range.end;
    const lineStart = lines[rangeStart.lineNumber];
    const lineEnd = lines[rangeEnd.lineNumber];
    if (rangeStart.lineCodeUnit > lineStart.text.length) {
        kind = _1.LexError.BadRangeKind.LineCodeUnitStart_GreaterThan_LineLength;
    }
    else if (rangeEnd.lineCodeUnit > lineEnd.text.length) {
        kind = _1.LexError.BadRangeKind.LineCodeUnitEnd_GreaterThan_LineLength;
    }
    if (kind !== undefined) {
        return new _1.LexError.BadRangeError(range, kind, state.locale);
    }
    return undefined;
}
//# sourceMappingURL=lexer.js.map