"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.trySnapshot = exports.LexerSnapshot = void 0;
const common_1 = require("../common");
const language_1 = require("../language");
const _1 = require(".");
class LexerSnapshot {
    constructor(text, tokens, comments, lineTerminators) {
        this.text = text;
        this.tokens = tokens;
        this.comments = comments;
        this.lineTerminators = lineTerminators;
    }
    static graphemePositionStartFrom(text, lineTerminators, flatLineToken) {
        const positionStart = flatLineToken.positionStart;
        const positionEnd = flatLineToken.positionEnd;
        let substringPositionStart = 0;
        let substringPositionEnd = text.length;
        for (const lineTerminator of lineTerminators) {
            if (lineTerminator.codeUnit < positionStart.codeUnit) {
                substringPositionStart = lineTerminator.codeUnit + lineTerminator.text.length;
            }
            if (lineTerminator.codeUnit >= positionEnd.codeUnit) {
                substringPositionEnd = lineTerminator.codeUnit + lineTerminator.text.length;
                break;
            }
        }
        return common_1.StringUtils.graphemePositionFrom(text.substring(substringPositionStart, substringPositionEnd), positionStart.lineCodeUnit, positionStart.lineNumber, positionEnd.codeUnit);
    }
    graphemePositionStartFrom(token) {
        return LexerSnapshot.graphemePositionStartFrom(this.text, this.lineTerminators, token);
    }
    columnNumberStartFrom(token) {
        return this.graphemePositionStartFrom(token).columnNumber;
    }
}
exports.LexerSnapshot = LexerSnapshot;
function trySnapshot(state) {
    try {
        return common_1.ResultUtils.ok(createSnapshot(state));
    }
    catch (caught) {
        let error;
        if (_1.LexError.isTInnerLexError(caught)) {
            error = new _1.LexError.LexError(caught);
        }
        else {
            common_1.Assert.isInstanceofError(caught);
            error = common_1.CommonError.ensureCommonError(caught, state.locale);
        }
        return common_1.ResultUtils.error(error);
    }
}
exports.trySnapshot = trySnapshot;
const ErrorMessageNoEndToken = "once a multiline token starts it should either reach a paired end token, or eof";
function createSnapshot(state) {
    var _a;
    // class properties
    const tokens = [];
    const comments = [];
    const flattenedLines = flattenLineTokens(state);
    const flatTokens = flattenedLines.flatLineTokens;
    const numFlatTokens = flatTokens.length;
    const text = flattenedLines.text;
    const cancellationToken = state.cancellationToken;
    let flatIndex = 0;
    while (flatIndex < numFlatTokens) {
        (_a = state.cancellationToken) === null || _a === void 0 ? void 0 : _a.throwIfCancelled();
        const flatToken = flatTokens[flatIndex];
        switch (flatToken.kind) {
            case language_1.Token.LineTokenKind.LineComment:
                comments.push(readLineComment(flatToken));
                break;
            case language_1.Token.LineTokenKind.MultilineComment:
                comments.push(readSingleLineMultilineComment(flatToken));
                break;
            case language_1.Token.LineTokenKind.MultilineCommentStart: {
                const concatenatedTokenRead = readMultilineComment(flattenedLines, flatToken, state.locale, cancellationToken);
                comments.push(concatenatedTokenRead.comment);
                flatIndex = concatenatedTokenRead.flatIndexEnd;
                break;
            }
            case language_1.Token.LineTokenKind.QuotedIdentifierStart: {
                const concatenatedTokenRead = readQuotedIdentifier(flattenedLines, flatToken, state.locale, cancellationToken);
                tokens.push(concatenatedTokenRead.token);
                flatIndex = concatenatedTokenRead.flatIndexEnd;
                break;
            }
            case language_1.Token.LineTokenKind.TextLiteralStart: {
                const concatenatedTokenRead = readTextLiteral(flattenedLines, flatToken, state.locale, cancellationToken);
                tokens.push(concatenatedTokenRead.token);
                flatIndex = concatenatedTokenRead.flatIndexEnd;
                break;
            }
            default: {
                const positionStart = flatToken.positionStart;
                const positionEnd = flatToken.positionEnd;
                tokens.push({
                    kind: flatToken.kind,
                    data: flatToken.data,
                    positionStart,
                    positionEnd,
                });
            }
        }
        flatIndex += 1;
    }
    return new LexerSnapshot(text, tokens, comments, flattenedLines.lineTerminators);
}
function readLineComment(flatToken) {
    const positionStart = flatToken.positionStart;
    const positionEnd = flatToken.positionEnd;
    return {
        kind: language_1.Comment.CommentKind.Line,
        data: flatToken.data,
        containsNewline: true,
        positionStart,
        positionEnd,
    };
}
// a multiline comment that spans a single line
function readSingleLineMultilineComment(flatToken) {
    const positionStart = flatToken.positionStart;
    const positionEnd = flatToken.positionEnd;
    return {
        kind: language_1.Comment.CommentKind.Multiline,
        data: flatToken.data,
        containsNewline: positionStart.lineNumber !== positionEnd.lineNumber,
        positionStart,
        positionEnd,
    };
}
function readMultilineComment(flattenedLines, tokenStart, locale, cancellationToken) {
    const collection = collectWhileContent(flattenedLines.flatLineTokens, tokenStart, language_1.Token.LineTokenKind.MultilineCommentContent, cancellationToken);
    const tokenEnd = collection.tokenEnd;
    if (!tokenEnd) {
        throw new _1.LexError.UnterminatedMultilineTokenError(locale, LexerSnapshot.graphemePositionStartFrom(flattenedLines.text, flattenedLines.lineTerminators, tokenStart), _1.LexError.UnterminatedMultilineTokenKind.MultilineComment);
    }
    else if (tokenEnd.kind !== language_1.Token.LineTokenKind.MultilineCommentEnd) {
        const details = { foundTokenEnd: tokenEnd };
        const message = ErrorMessageNoEndToken;
        throw new common_1.CommonError.InvariantError(message, details);
    }
    else {
        const positionStart = tokenStart.positionStart;
        const positionEnd = tokenEnd.positionEnd;
        return {
            comment: {
                kind: language_1.Comment.CommentKind.Multiline,
                data: flattenedLines.text.substring(positionStart.codeUnit, positionEnd.codeUnit),
                containsNewline: positionStart.lineNumber !== positionEnd.lineNumber,
                positionStart,
                positionEnd,
            },
            flatIndexEnd: tokenEnd.flatIndex,
        };
    }
}
function readQuotedIdentifier(flattenedLines, tokenStart, locale, cancellationToken) {
    const collection = collectWhileContent(flattenedLines.flatLineTokens, tokenStart, language_1.Token.LineTokenKind.QuotedIdentifierContent, cancellationToken);
    const tokenEnd = collection.tokenEnd;
    if (!tokenEnd) {
        throw new _1.LexError.UnterminatedMultilineTokenError(locale, LexerSnapshot.graphemePositionStartFrom(flattenedLines.text, flattenedLines.lineTerminators, tokenStart), _1.LexError.UnterminatedMultilineTokenKind.QuotedIdentifier);
    }
    else if (tokenEnd.kind !== language_1.Token.LineTokenKind.QuotedIdentifierEnd) {
        const details = { foundTokenEnd: tokenEnd };
        const message = ErrorMessageNoEndToken;
        throw new common_1.CommonError.InvariantError(message, details);
    }
    else {
        const positionStart = tokenStart.positionStart;
        const positionEnd = tokenEnd.positionEnd;
        return {
            token: {
                kind: language_1.Token.TokenKind.Identifier,
                data: flattenedLines.text.substring(positionStart.codeUnit, positionEnd.codeUnit),
                positionStart,
                positionEnd,
            },
            flatIndexEnd: tokenEnd.flatIndex,
        };
    }
}
function readTextLiteral(flattenedLines, tokenStart, locale, cancellationToken) {
    const collection = collectWhileContent(flattenedLines.flatLineTokens, tokenStart, language_1.Token.LineTokenKind.TextLiteralContent, cancellationToken);
    const tokenEnd = collection.tokenEnd;
    if (!tokenEnd) {
        throw new _1.LexError.UnterminatedMultilineTokenError(locale, LexerSnapshot.graphemePositionStartFrom(flattenedLines.text, flattenedLines.lineTerminators, tokenStart), _1.LexError.UnterminatedMultilineTokenKind.Text);
    }
    else if (tokenEnd.kind !== language_1.Token.LineTokenKind.TextLiteralEnd) {
        const details = { foundTokenEnd: tokenEnd };
        const message = ErrorMessageNoEndToken;
        throw new common_1.CommonError.InvariantError(message, details);
    }
    else {
        const positionStart = tokenStart.positionStart;
        const positionEnd = tokenEnd.positionEnd;
        return {
            token: {
                kind: language_1.Token.TokenKind.TextLiteral,
                data: flattenedLines.text.substring(positionStart.codeUnit, positionEnd.codeUnit),
                positionStart,
                positionEnd,
            },
            flatIndexEnd: tokenEnd.flatIndex,
        };
    }
}
function collectWhileContent(flatTokens, tokenStart, contentKind, cancellationToken) {
    const collectedTokens = [];
    const numTokens = flatTokens.length;
    let flatIndex = tokenStart.flatIndex + 1;
    while (flatIndex < numTokens) {
        cancellationToken === null || cancellationToken === void 0 ? void 0 : cancellationToken.throwIfCancelled();
        const token = flatTokens[flatIndex];
        if (token.kind !== contentKind) {
            break;
        }
        collectedTokens.push(token);
        flatIndex += 1;
    }
    return {
        tokenStart,
        collectedTokens,
        tokenEnd: flatTokens[flatIndex],
    };
}
function flattenLineTokens(state) {
    const lines = state.lines;
    const lineTerminators = [];
    const numLines = lines.length;
    let text = "";
    const flatLineTokens = [];
    let lineTextOffset = 0;
    let flatIndex = 0;
    for (const [line, lineNumber] of common_1.ArrayUtils.enumerate(lines)) {
        text += line.text;
        if (lineNumber !== numLines - 1) {
            text += line.lineTerminator;
        }
        for (const lineToken of line.tokens) {
            const linePositionStart = lineToken.positionStart;
            const linePositionEnd = lineToken.positionEnd;
            flatLineTokens.push({
                kind: lineToken.kind,
                data: lineToken.data,
                positionStart: {
                    codeUnit: lineTextOffset + linePositionStart,
                    lineCodeUnit: linePositionStart,
                    lineNumber,
                },
                positionEnd: {
                    codeUnit: lineTextOffset + linePositionEnd,
                    lineCodeUnit: linePositionEnd,
                    lineNumber,
                },
                flatIndex,
            });
            flatIndex += 1;
        }
        const lineTerminatorCodeUnit = lineTextOffset + line.text.length;
        lineTerminators.push({
            codeUnit: lineTerminatorCodeUnit,
            text: line.lineTerminator,
        });
        lineTextOffset = lineTerminatorCodeUnit + line.lineTerminator.length;
    }
    return {
        text,
        lineTerminators,
        flatLineTokens,
    };
}
//# sourceMappingURL=lexerSnapshot.js.map