"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.deleteContext = exports.deleteAst = exports.endContext = exports.startContextAsParent = exports.startContext = exports.nextAttributeIndex = exports.nextId = exports.isNodeKind = exports.copyState = exports.newState = exports.assertIsNodeKind = exports.assertAsNodeKind = void 0;
const common_1 = require("../../common");
const nodeIdMap_1 = require("../nodeIdMap");
function assertAsNodeKind(node, expectedNodeKinds) {
    assertIsNodeKind(node, expectedNodeKinds);
    return node;
}
exports.assertAsNodeKind = assertAsNodeKind;
function assertIsNodeKind(node, expectedNodeKinds) {
    if (!isNodeKind(node, expectedNodeKinds)) {
        throw new common_1.CommonError.InvariantError(`expected parse context node has a different than expected node kind`, {
            nodeId: node.id,
            nodeKind: node.kind,
            expectedNodeKinds,
        });
    }
}
exports.assertIsNodeKind = assertIsNodeKind;
function newState() {
    return {
        nodeIdMapCollection: {
            astNodeById: new Map(),
            childIdsById: new Map(),
            contextNodeById: new Map(),
            leafIds: new Set(),
            rightMostLeaf: undefined,
            idsByNodeKind: new Map(),
            parentIdById: new Map(),
        },
        root: undefined,
        idCounter: 0,
    };
}
exports.newState = newState;
function copyState(state) {
    const root = state.root !== undefined ? Object.assign({}, state.root) : undefined;
    return Object.assign(Object.assign({}, state), { root, nodeIdMapCollection: nodeIdMap_1.NodeIdMapUtils.copy(state.nodeIdMapCollection) });
}
exports.copyState = copyState;
function isNodeKind(node, expectedNodeKinds) {
    return node.kind === expectedNodeKinds || expectedNodeKinds.includes(node.kind);
}
exports.isNodeKind = isNodeKind;
function nextId(state) {
    state.idCounter += 1;
    return state.idCounter;
}
exports.nextId = nextId;
function nextAttributeIndex(parentNode) {
    const result = parentNode.attributeCounter;
    parentNode.attributeCounter += 1;
    return result;
}
exports.nextAttributeIndex = nextAttributeIndex;
function startContext(state, nodeKind, tokenIndexStart, tokenStart, parentNode) {
    const nodeIdMapCollection = state.nodeIdMapCollection;
    let attributeIndex;
    const nodeId = nextId(state);
    // If a parent context Node exists, update the parent/child mapping attributes and attributeCounter.
    if (parentNode) {
        const childIdsById = nodeIdMapCollection.childIdsById;
        const parentId = parentNode.id;
        attributeIndex = nextAttributeIndex(parentNode);
        nodeIdMapCollection.parentIdById.set(nodeId, parentId);
        const existingChildren = childIdsById.get(parentId);
        if (existingChildren) {
            childIdsById.set(parentId, [...existingChildren, nodeId]);
        }
        else {
            childIdsById.set(parentId, [nodeId]);
        }
    }
    const contextNode = {
        id: nodeId,
        kind: nodeKind,
        tokenIndexStart,
        tokenStart,
        attributeCounter: 0,
        isClosed: false,
        attributeIndex,
    };
    nodeIdMapCollection.contextNodeById.set(nodeId, contextNode);
    if (state.root === undefined) {
        state.root = contextNode;
    }
    trackNodeIdByNodeKind(nodeIdMapCollection.idsByNodeKind, nodeKind, nodeId);
    return contextNode;
}
exports.startContext = startContext;
function startContextAsParent(state, nodeKind, existingNodeId, existingNodeTokenStart) {
    const nodeIdMapCollection = state.nodeIdMapCollection;
    const existingNode = nodeIdMap_1.NodeIdMapUtils.assertXor(nodeIdMapCollection, existingNodeId);
    let tokenIndexStart;
    let attributeIndex;
    // Update `attributeIndex` as we're inserting a new context as the parent of the existing node.
    switch (existingNode.kind) {
        case nodeIdMap_1.XorNodeKind.Ast: {
            tokenIndexStart = existingNode.node.tokenRange.tokenIndexStart;
            attributeIndex = existingNode.node.attributeIndex;
            const mutableAst = existingNode.node;
            mutableAst.attributeIndex = 0;
            break;
        }
        case nodeIdMap_1.XorNodeKind.Context:
            tokenIndexStart = existingNode.node.tokenIndexStart;
            attributeIndex = existingNode.node.attributeIndex;
            existingNode.node.attributeIndex = 0;
            break;
        default:
            throw common_1.Assert.isNever(existingNode);
    }
    const newNodeId = nextId(state);
    const insertedContext = {
        id: newNodeId,
        kind: nodeKind,
        tokenIndexStart,
        tokenStart: existingNodeTokenStart,
        // We want to create a new context containing an existing node,
        // so attributeCounter must be 1 as it holds the existing node.
        attributeCounter: 1,
        isClosed: false,
        attributeIndex,
    };
    nodeIdMapCollection.contextNodeById.set(newNodeId, insertedContext);
    trackNodeIdByNodeKind(nodeIdMapCollection.idsByNodeKind, nodeKind, newNodeId);
    // We have one of two possible states, either:
    //  1) grandparent <-> new <-> existing
    //  2) new <-> existing.
    // Scenario (1)
    const grandparentNodeId = nodeIdMapCollection.parentIdById.get(existingNodeId);
    if (grandparentNodeId) {
        // Update `grandparent <-> existing` with `grandparent <-> new`
        removeOrReplaceChildId(nodeIdMapCollection, grandparentNodeId, existingNodeId, newNodeId);
    }
    // Applicable to Scenario (1) and (2)
    // Create `new <-> existing`
    nodeIdMapCollection.parentIdById.set(existingNodeId, newNodeId);
    nodeIdMapCollection.childIdsById.set(newNodeId, [existingNodeId]);
    return insertedContext;
}
exports.startContextAsParent = startContextAsParent;
// Returns the Node's parent context (if one exists).
function endContext(state, contextNode, astNode) {
    var _a;
    const nodeIdMapCollection = state.nodeIdMapCollection;
    if (((_a = state.root) === null || _a === void 0 ? void 0 : _a.id) === astNode.id) {
        common_1.Assert.isTrue(state.nodeIdMapCollection.contextNodeById.size === 1, "failed to endContext on the root node as at least one other context node exists", { nodeId: contextNode.id });
    }
    contextNode.isClosed = true;
    if (astNode.isLeaf) {
        common_1.SetUtils.assertAddUnique(state.nodeIdMapCollection.leafIds, contextNode.id, `failed to endContext on a leaf node as it shares a nodeId with an already existing leaf node`, { nodeId: contextNode.id });
    }
    // Ending a context should return the context's parent node (if one exists).
    const parentId = nodeIdMapCollection.parentIdById.get(contextNode.id);
    const parentNode = parentId !== undefined ? nodeIdMapCollection.contextNodeById.get(parentId) : undefined;
    // Move nodeId from contextNodeMap to astNodeMap.
    common_1.MapUtils.assertDelete(nodeIdMapCollection.contextNodeById, contextNode.id, "failed to endContext as the given nodeId isn't a valid context nodeId", { nodeId: contextNode.id });
    nodeIdMapCollection.astNodeById.set(astNode.id, astNode);
    // Update rightMostLeaf when applicable
    if (astNode.isLeaf) {
        if (nodeIdMapCollection.rightMostLeaf === undefined ||
            nodeIdMapCollection.rightMostLeaf.tokenRange.tokenIndexStart < astNode.tokenRange.tokenIndexStart) {
            const unsafeNodeIdMapCollection = nodeIdMapCollection;
            unsafeNodeIdMapCollection.rightMostLeaf = astNode;
        }
    }
    return parentNode;
}
exports.endContext = endContext;
function deleteAst(state, nodeId, parentWillBeDeleted) {
    const nodeIdMapCollection = state.nodeIdMapCollection;
    const astNodeById = nodeIdMapCollection.astNodeById;
    const astNode = common_1.MapUtils.assertGet(astNodeById, nodeId, `failed to deleteAst as the given nodeId isn't a valid Ast nodeId`, { nodeId });
    const childIdsById = nodeIdMapCollection.childIdsById;
    const parentIdById = nodeIdMapCollection.parentIdById;
    // If Node was a leaf node, remove it from the list of leaf nodes.
    if (astNode.isLeaf) {
        common_1.SetUtils.assertDelete(nodeIdMapCollection.leafIds, nodeId, "failed to deleteAst as the node is a leaf node, but it wasn't present in leafIds", { nodeId });
    }
    const parentId = parentIdById.get(nodeId);
    const childIds = childIdsById.get(nodeId);
    // Not a leaf node.
    common_1.Assert.isUndefined(childIds, `failed to deleteAst as the given nodeId has one ore more children which must be deleted first`, {
        nodeId,
        childIds,
    });
    // Is a leaf node, not root node.
    // Delete the node from the list of children under the node's parent.
    if (parentId) {
        common_1.Assert.isFalse(astNodeById.has(parentId) && !parentWillBeDeleted, `parent is an Ast node not marked for deletion`, {
            parentId,
            nodeId,
        });
        removeOrReplaceChildId(nodeIdMapCollection, parentId, nodeId, undefined);
    }
    // Else is root node, is leaf node.
    // No children updates need to be taken.
    // Remove the node from existence.
    deleteFromKindMap(nodeIdMapCollection, nodeId);
    astNodeById.delete(nodeId);
    childIdsById.delete(nodeId);
    parentIdById.delete(nodeId);
}
exports.deleteAst = deleteAst;
function deleteContext(state, nodeId) {
    const nodeIdMapCollection = state.nodeIdMapCollection;
    const childIdsById = nodeIdMapCollection.childIdsById;
    const contextNodeById = nodeIdMapCollection.contextNodeById;
    const leafIds = nodeIdMapCollection.leafIds;
    const parentIdById = nodeIdMapCollection.parentIdById;
    const contextNode = common_1.MapUtils.assertGet(contextNodeById, nodeId, `failed to deleteContext as the given nodeId isn't a valid context node`, { nodeId });
    const parentId = parentIdById.get(nodeId);
    const childIds = childIdsById.get(nodeId);
    // Not a leaf node.
    if (childIds !== undefined) {
        common_1.ArrayUtils.assertNonZeroLength(childIds);
        const childId = childIds[0];
        // Not a leaf node, is the Root node.
        // Promote the child to the root if it's a Context node.
        if (parentId === undefined) {
            parentIdById.delete(childId);
            const childContext = contextNodeById.get(childId);
            if (childContext) {
                state.root = childContext;
            }
        }
        // Not a leaf node, not the Root node.
        // Replace the node from the list of children under the node's parent using the node's child
        else {
            removeOrReplaceChildId(nodeIdMapCollection, parentId, nodeId, childId);
        }
        // The child Node inherits the attributeIndex.
        const childXorNode = nodeIdMap_1.NodeIdMapUtils.assertXor(state.nodeIdMapCollection, childId);
        const mutableChildXorNode = childXorNode.node;
        mutableChildXorNode.attributeIndex = contextNode.attributeIndex;
    }
    // Is a leaf node, not root node.
    // Delete the node from the list of children under the node's parent.
    else if (parentId) {
        removeOrReplaceChildId(nodeIdMapCollection, parentId, nodeId, undefined);
    }
    // Else is root node, is leaf node.
    // No children updates need to be taken.
    // Remove the node from existence.
    deleteFromKindMap(nodeIdMapCollection, nodeId);
    contextNodeById.delete(nodeId);
    childIdsById.delete(nodeId);
    parentIdById.delete(nodeId);
    leafIds.delete(nodeId);
    // Return the node's parent if it exits
    return parentId !== undefined ? nodeIdMap_1.NodeIdMapUtils.assertContext(nodeIdMapCollection, parentId) : undefined;
}
exports.deleteContext = deleteContext;
function trackNodeIdByNodeKind(idsByNodeKind, nodeKind, nodeId) {
    const idsForSpecificNodeKind = idsByNodeKind.get(nodeKind);
    if (idsForSpecificNodeKind) {
        idsForSpecificNodeKind.add(nodeId);
    }
    else {
        idsByNodeKind.set(nodeKind, new Set([nodeId]));
    }
}
function deleteFromKindMap(nodeIdMapCollection, nodeId) {
    const idsByNodeKind = nodeIdMapCollection.idsByNodeKind;
    const nodeKind = nodeIdMap_1.NodeIdMapUtils.assertXor(nodeIdMapCollection, nodeId).node.kind;
    const nodeIds = common_1.MapUtils.assertGet(idsByNodeKind, nodeKind, `failed to deleteFromKindMap as the node kind for the given nodeId isn't present in idsByNodeKind`, {
        nodeId,
        nodeKind,
    });
    common_1.SetUtils.assertDelete(nodeIds, nodeId);
    if (!nodeIds.size) {
        idsByNodeKind.delete(nodeKind);
    }
}
// Asserts `childId` is in the list of children in for `parentId`.
// Either removes `childId` in its list of children if no `replcementId` is given,
// else replaces the `childId` with `replacementId`.
function removeOrReplaceChildId(nodeIdMapCollection, parentId, childId, replacementId) {
    const childIdsById = nodeIdMapCollection.childIdsById;
    const childIds = nodeIdMap_1.NodeIdMapUtils.assertChildIds(childIdsById, parentId);
    let newChildIds;
    if (replacementId) {
        nodeIdMapCollection.parentIdById.set(replacementId, parentId);
        newChildIds = common_1.ArrayUtils.assertReplaceFirstInstance(childIds, childId, replacementId);
    }
    else {
        newChildIds = common_1.ArrayUtils.assertRemoveFirstInstance(childIds, childId);
    }
    if (newChildIds.length) {
        childIdsById.set(parentId, newChildIds);
    }
    else {
        childIdsById.delete(parentId);
    }
    const parent = nodeIdMapCollection.contextNodeById.get(parentId);
    if (parent !== undefined && replacementId === undefined) {
        parent.attributeCounter -= 1;
    }
}
//# sourceMappingURL=contextUtils.js.map