"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.disambiguateBracket = exports.disambiguateParenthesis = exports.readAmbiguousParenthesis = exports.readAmbiguousBracket = exports.readAmbiguous = void 0;
const disambiguation_1 = require("./disambiguation");
const common_1 = require("../../common");
const language_1 = require("../../language");
const parseState_1 = require("../parseState");
const trace_1 = require("../../common/trace");
const nodeIdMap_1 = require("../nodeIdMap");
const __1 = require("..");
// For each given parse function it'll create a deep copy of the state then parse with the function.
// Mutates the given state to whatever parse state which matched the most amount of tokens.
// Ties are resolved by the ordering of the given parse functions.
function readAmbiguous(state, parser, parseCallbacks, correlationId) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = state.traceManager.entry("Disambiguation" /* Disambiguation */, readAmbiguous.name, correlationId, {
            [trace_1.TraceConstant.Length]: parseCallbacks.length,
        });
        common_1.ArrayUtils.assertNonZeroLength(parseCallbacks, "requires at least one parse function");
        let bestMatch = undefined;
        for (const parseCallback of parseCallbacks) {
            // eslint-disable-next-line no-await-in-loop
            const variantState = yield parser.copyState(state);
            let node;
            let variantResult;
            try {
                // eslint-disable-next-line no-await-in-loop
                node = yield parseCallback(variantState, parser, trace.id);
                variantResult = common_1.ResultUtils.ok(node);
            }
            catch (error) {
                common_1.Assert.isInstanceofError(error);
                common_1.CommonError.throwIfCancellationError(error);
                if (!__1.ParseError.isTInnerParseError(error)) {
                    throw error;
                }
                variantResult = common_1.ResultUtils.error(new __1.ParseError.ParseError(error, variantState));
            }
            const candiate = {
                parseState: variantState,
                result: variantResult,
            };
            bestMatch = bestAmbiguousParseMatch(bestMatch, candiate);
        }
        common_1.Assert.isDefined(bestMatch);
        trace.exit();
        return bestMatch;
    });
}
exports.readAmbiguous = readAmbiguous;
// Peeks at the token stream and either performs an explicit read or an ambiguous read.
function readAmbiguousBracket(state, parser, allowedVariants, correlationId) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = state.traceManager.entry("Disambiguation" /* Disambiguation */, readAmbiguousBracket.name, correlationId);
        // We might be able to peek at tokens to disambiguate what bracketed expression is next.
        const disambiguation = disambiguateBracket(state, allowedVariants, trace.id);
        // Peeking gave us a concrete answer as to what's next.
        let ambiguousBracket;
        if (disambiguation !== undefined) {
            switch (disambiguation) {
                case disambiguation_1.BracketDisambiguation.FieldProjection:
                    ambiguousBracket = yield parser.readFieldProjection(state, parser, trace.id);
                    break;
                case disambiguation_1.BracketDisambiguation.FieldSelection:
                    ambiguousBracket = yield parser.readFieldSelection(state, parser, trace.id);
                    break;
                case disambiguation_1.BracketDisambiguation.RecordExpression:
                    ambiguousBracket = yield parser.readRecordExpression(state, parser, trace.id);
                    break;
                default:
                    throw common_1.Assert.isNever(disambiguation);
            }
        }
        // Else we branch on `IParseState.disambiguousBehavior`.
        else {
            switch (state.disambiguationBehavior) {
                case disambiguation_1.DismabiguationBehavior.Strict:
                    trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                    throw parseState_1.ParseStateUtils.unterminatedBracketError(state);
                case disambiguation_1.DismabiguationBehavior.Thorough:
                    ambiguousBracket = yield thoroughReadAmbiguousBracket(state, parser, allowedVariants, trace.id);
                    break;
                default:
                    throw common_1.Assert.isNever(state.disambiguationBehavior);
            }
        }
        trace.exit({ [trace_1.TraceConstant.IsThrowing]: false });
        return ambiguousBracket;
    });
}
exports.readAmbiguousBracket = readAmbiguousBracket;
// Peeks at the token stream and either performs an explicit read or an ambiguous read.
function readAmbiguousParenthesis(state, parser, correlationId) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = state.traceManager.entry("Disambiguation" /* Disambiguation */, readAmbiguousParenthesis.name, correlationId);
        // We might be able to peek at tokens to disambiguate what parenthesized expression is next.
        const disambiguation = yield disambiguateParenthesis(state, parser, trace.id);
        // Peeking gave us a concrete answer as to what's next.
        let ambiguousParenthesis;
        if (disambiguation !== undefined) {
            switch (disambiguation) {
                case disambiguation_1.ParenthesisDisambiguation.FunctionExpression:
                    ambiguousParenthesis = yield parser.readFunctionExpression(state, parser, trace.id);
                    break;
                case disambiguation_1.ParenthesisDisambiguation.ParenthesizedExpression:
                    ambiguousParenthesis = yield readParenthesizedExpressionOrBinOpExpression(state, parser, trace.id);
                    break;
                default:
                    throw common_1.Assert.isNever(disambiguation);
            }
        }
        // Else we branch on `IParseState.disambiguousBehavior`.
        else {
            switch (state.disambiguationBehavior) {
                case disambiguation_1.DismabiguationBehavior.Strict:
                    trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                    throw parseState_1.ParseStateUtils.unterminatedParenthesesError(state);
                case disambiguation_1.DismabiguationBehavior.Thorough:
                    ambiguousParenthesis = yield thoroughReadAmbiguousParenthesis(state, parser, trace.id);
                    break;
                default:
                    trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
                    throw common_1.Assert.isNever(state.disambiguationBehavior);
            }
        }
        trace.exit({ [trace_1.TraceConstant.IsThrowing]: false });
        return ambiguousParenthesis;
    });
}
exports.readAmbiguousParenthesis = readAmbiguousParenthesis;
// Peeks at tokens which might give a concrete disambiguation.
function disambiguateParenthesis(state, parser, correlationId) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = state.traceManager.entry("Disambiguation" /* Disambiguation */, disambiguateParenthesis.name, correlationId);
        const initialTokenIndex = state.tokenIndex;
        const tokens = state.lexerSnapshot.tokens;
        const totalTokens = tokens.length;
        let nestedDepth = 1;
        let offsetTokenIndex = initialTokenIndex + 1;
        while (offsetTokenIndex < totalTokens) {
            const offsetTokenKind = tokens[offsetTokenIndex].kind;
            if (offsetTokenKind === language_1.Token.TokenKind.LeftParenthesis) {
                nestedDepth += 1;
            }
            else if (offsetTokenKind === language_1.Token.TokenKind.RightParenthesis) {
                nestedDepth -= 1;
            }
            let disambiguation = undefined;
            if (nestedDepth === 0) {
                // '(x as number) as number' could either be either case,
                // so we need to consume test if the trailing 'as number' is followed by a FatArrow.
                if (parseState_1.ParseStateUtils.isTokenKind(state, language_1.Token.TokenKind.KeywordAs, offsetTokenIndex + 1)) {
                    // eslint-disable-next-line no-await-in-loop
                    const checkpoint = yield parser.checkpoint(state);
                    unsafeMoveTo(state, offsetTokenIndex + 2);
                    try {
                        // eslint-disable-next-line no-await-in-loop
                        yield parser.readNullablePrimitiveType(state, parser, trace.id);
                    }
                    catch (_a) {
                        // eslint-disable-next-line no-await-in-loop
                        yield parser.restoreCheckpoint(state, checkpoint);
                        if (parseState_1.ParseStateUtils.isOnTokenKind(state, language_1.Token.TokenKind.FatArrow)) {
                            return disambiguation_1.ParenthesisDisambiguation.FunctionExpression;
                        }
                        else {
                            return disambiguation_1.ParenthesisDisambiguation.ParenthesizedExpression;
                        }
                    }
                    if (parseState_1.ParseStateUtils.isOnTokenKind(state, language_1.Token.TokenKind.FatArrow)) {
                        disambiguation = disambiguation_1.ParenthesisDisambiguation.FunctionExpression;
                    }
                    else {
                        disambiguation = disambiguation_1.ParenthesisDisambiguation.ParenthesizedExpression;
                    }
                    // eslint-disable-next-line no-await-in-loop
                    yield parser.restoreCheckpoint(state, checkpoint);
                }
                else if (parseState_1.ParseStateUtils.isTokenKind(state, language_1.Token.TokenKind.FatArrow, offsetTokenIndex + 1)) {
                    disambiguation = disambiguation_1.ParenthesisDisambiguation.FunctionExpression;
                }
                else {
                    disambiguation = disambiguation_1.ParenthesisDisambiguation.ParenthesizedExpression;
                }
            }
            if (disambiguation) {
                trace.exit({ [trace_1.TraceConstant.Result]: disambiguation });
                return disambiguation;
            }
            offsetTokenIndex += 1;
        }
        trace.exit({ [trace_1.TraceConstant.Result]: undefined });
        return undefined;
    });
}
exports.disambiguateParenthesis = disambiguateParenthesis;
function disambiguateBracket(state, allowedVariants, correlationId) {
    const trace = state.traceManager.entry("Disambiguation" /* Disambiguation */, disambiguateBracket.name, correlationId);
    let offsetTokenIndex = state.tokenIndex + 1;
    const tokens = state.lexerSnapshot.tokens;
    const offsetToken = tokens[offsetTokenIndex];
    if (offsetToken === undefined) {
        return undefined;
    }
    let offsetTokenKind = offsetToken.kind;
    let result;
    if (offsetTokenKind === language_1.Token.TokenKind.LeftBracket) {
        result = disambiguation_1.BracketDisambiguation.FieldProjection;
    }
    else if (offsetTokenKind === language_1.Token.TokenKind.RightBracket) {
        result = disambiguation_1.BracketDisambiguation.RecordExpression;
    }
    else {
        const totalTokens = tokens.length;
        offsetTokenIndex += 1;
        while (offsetTokenIndex < totalTokens) {
            offsetTokenKind = tokens[offsetTokenIndex].kind;
            if (offsetTokenKind === language_1.Token.TokenKind.Equal) {
                result = disambiguation_1.BracketDisambiguation.RecordExpression;
                break;
            }
            else if (offsetTokenKind === language_1.Token.TokenKind.RightBracket) {
                result = disambiguation_1.BracketDisambiguation.FieldSelection;
                break;
            }
            offsetTokenIndex += 1;
        }
    }
    trace.exit({ [trace_1.TraceConstant.Result]: result });
    return result !== undefined && allowedVariants.includes(result) ? result : undefined;
}
exports.disambiguateBracket = disambiguateBracket;
// Copy the current state and attempt to read for each of the following:
//  FieldProjection, FieldSelection, and RecordExpression.
// Mutates the given state with the read attempt which matched the most tokens.
function thoroughReadAmbiguousBracket(state, parser, allowedVariants, correlationId) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = state.traceManager.entry("Disambiguation" /* Disambiguation */, readAmbiguousBracket.name, correlationId);
        const ambiguousBracket = yield thoroughReadAmbiguous(state, parser, bracketDisambiguationParseFunctions(parser, allowedVariants), trace.id);
        trace.exit({ allowedVariants });
        return ambiguousBracket;
    });
}
// Copy the current state and attempt to read for each of the following:
//  FunctionExpression, ParenthesisExpression.
// Mutates the given state with the read attempt which matched the most tokens.
function thoroughReadAmbiguousParenthesis(state, parser, correlationId) {
    return thoroughReadAmbiguous(state, parser, [parser.readFunctionExpression, readParenthesizedExpressionOrBinOpExpression], correlationId);
}
function thoroughReadAmbiguous(state, parser, parseFns, correlationId) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = state.traceManager.entry("Disambiguation" /* Disambiguation */, thoroughReadAmbiguous.name, correlationId, {
            [trace_1.TraceConstant.Length]: parseFns.length,
        });
        const ambiguousParse = yield readAmbiguous(state, parser, parseFns, trace.id);
        yield parser.applyState(state, ambiguousParse.parseState);
        if (common_1.ResultUtils.isOk(ambiguousParse.result)) {
            trace.exit({
                [trace_1.TraceConstant.IsThrowing]: false,
                [trace_1.TraceConstant.Result]: ambiguousParse.result,
            });
            return ambiguousParse.result.value;
        }
        else {
            // ParseError.state references the cloned state generated in readAmbiguous, not the current state parameter.
            // For correctness sake we need to update the existing state with the AmbiguousParse error,
            // then update the reference.
            const mutableParseError = ambiguousParse.result.error;
            mutableParseError.state = state;
            trace.exit({ [trace_1.TraceConstant.IsThrowing]: true });
            throw ambiguousParse.result.error;
        }
    });
}
// Converts BracketDisambiguation into its corrosponding read function.
function bracketDisambiguationParseFunctions(parser, allowedVariants) {
    return allowedVariants.map((bracketDisambiguation) => {
        switch (bracketDisambiguation) {
            case disambiguation_1.BracketDisambiguation.FieldProjection:
                return parser.readFieldProjection;
            case disambiguation_1.BracketDisambiguation.FieldSelection:
                return parser.readFieldSelection;
            case disambiguation_1.BracketDisambiguation.RecordExpression:
                return parser.readRecordExpression;
            default:
                throw common_1.Assert.isNever(bracketDisambiguation);
        }
    });
}
// When the next token is an open parenthesis we can't directly read
// a ParenthesisExpression as it may leave trailing tokens behind.
// `(1) + 2`
function readParenthesizedExpressionOrBinOpExpression(state, parser, correlationId) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = state.traceManager.entry("Disambiguation" /* Disambiguation */, readParenthesizedExpressionOrBinOpExpression.name, correlationId);
        const node = yield parser.readLogicalExpression(state, parser, trace.id);
        const leftMostNode = nodeIdMap_1.NodeIdMapUtils.assertLeftMostLeaf(state.contextState.nodeIdMapCollection, node.id);
        language_1.AstUtils.assertAsNodeKind(leftMostNode, language_1.Ast.NodeKind.Constant);
        common_1.Assert.isTrue(leftMostNode.kind === language_1.Ast.NodeKind.Constant &&
            leftMostNode.constantKind === language_1.Constant.WrapperConstant.LeftParenthesis, `leftMostNode should be a ${language_1.Ast.NodeKind.Constant} with a constantKind of ${language_1.Constant.WrapperConstant.LeftParenthesis}`);
        trace.exit();
        return node;
    });
}
// WARNING: Only updates tokenIndex and currentTokenKind,
//          Manual cleanup of other state fields such as TokenRangeStack is assumed by the caller.
function unsafeMoveTo(state, tokenIndex) {
    const tokens = state.lexerSnapshot.tokens;
    state.tokenIndex = tokenIndex;
    if (tokenIndex < tokens.length) {
        state.currentToken = tokens[tokenIndex];
        state.currentTokenKind = state.currentToken.kind;
    }
    else {
        state.currentToken = undefined;
        state.currentTokenKind = undefined;
    }
}
function bestAmbiguousParseMatch(best, candidate) {
    if (best === undefined || best.parseState.tokenIndex < candidate.parseState.tokenIndex) {
        return candidate;
    }
    else if (best.parseState.tokenIndex === candidate.parseState.tokenIndex &&
        common_1.ResultUtils.isError(best.result) &&
        common_1.ResultUtils.isOk(candidate.result)) {
        return candidate;
    }
    else {
        return best;
    }
}
//# sourceMappingURL=disambiguationUtils.js.map