"use strict";
/* eslint-disable @typescript-eslint/no-unused-vars */
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.updateNodeIds = exports.recalculateIds = exports.recalculateAndUpdateIds = void 0;
const common_1 = require("../../../common");
const trace_1 = require("../../../common/trace");
// Helper functions which are related to updating / remapping nodeIds for NodeIdMap.Collection
function recalculateAndUpdateIds(nodeIdMapCollection, nodeId, traceManager, correlationId) {
    const trace = traceManager.entry("IdUtils" /* IdUtils */, recalculateAndUpdateIds.name, correlationId);
    const newIdByOldId = recalculateIds(nodeIdMapCollection, nodeId, traceManager, trace.id);
    updateNodeIds(nodeIdMapCollection, newIdByOldId, traceManager, trace.id);
    trace.exit();
}
exports.recalculateAndUpdateIds = recalculateAndUpdateIds;
// Builds up a list of all nodeIds under the given nodeId (including itself),
// then creates a Map<oldId, newId> such that the Ids are in a BFS ordering.
// Does not include nodeIds which remain unchanged.
function recalculateIds(nodeIdMapCollection, nodeId, traceManager, correlationId) {
    const trace = traceManager.entry("IdUtils" /* IdUtils */, recalculateIds.name, correlationId);
    const encounteredIds = [];
    let currentId = nodeId;
    let idQueue = [];
    while (currentId) {
        encounteredIds.push(currentId);
        const childIdsOfCurrentNode = nodeIdMapCollection.childIdsById.get(currentId);
        if (childIdsOfCurrentNode) {
            idQueue = childIdsOfCurrentNode.concat(idQueue);
        }
        currentId = idQueue.shift();
    }
    const numIds = encounteredIds.length;
    const sortedIds = [...encounteredIds].sort();
    const newIdByOldId = new Map();
    for (let index = 0; index < numIds; index += 1) {
        const oldId = encounteredIds[index];
        const newId = sortedIds[index];
        if (oldId !== newId) {
            newIdByOldId.set(oldId, newId);
        }
    }
    trace.exit();
    return newIdByOldId;
}
exports.recalculateIds = recalculateIds;
// Given a mapping of (existingId) => (newId) this mutates the NodeIdMap.Collection and the TXorNodes it holds.
// Assumes the given arguments are valid as this function does no validation.
function updateNodeIds(nodeIdMapCollection, newIdByOldId, traceManager, correlationId) {
    const trace = traceManager.entry("IdUtils" /* IdUtils */, updateNodeIds.name, correlationId);
    if (newIdByOldId.size === 0) {
        trace.exit({ [trace_1.TraceConstant.Size]: newIdByOldId.size });
        return;
    }
    // Storage for the change delta which is used to mutate nodeIdMapCollection.
    const collectionDelta = createDelta(nodeIdMapCollection, newIdByOldId, traceManager, trace.id);
    applyCollectionDelta(nodeIdMapCollection, collectionDelta, traceManager, trace.id);
    trace.exit({ [trace_1.TraceConstant.Size]: newIdByOldId.size });
}
exports.updateNodeIds = updateNodeIds;
function updatedChildIds(newIdByOldId, oldChildIds) {
    return oldChildIds.map((childId) => { var _a; return (_a = newIdByOldId.get(childId)) !== null && _a !== void 0 ? _a : childId; });
}
function createDelta(nodeIdMapCollection, newIdByOldId, traceManager, correlationId) {
    var _a, _b, _c;
    const trace = traceManager.entry("IdUtils" /* IdUtils */, createDelta.name, correlationId, {
        [trace_1.TraceConstant.Size]: newIdByOldId.size,
    });
    const collectionDelta = {
        astNodeById: new Map(),
        contextNodeById: new Map(),
        parentIdById: new Map(),
        childIdsById: new Map(),
        idsByNodeKind: new Map(),
        leafIds: new Set([...nodeIdMapCollection.leafIds].map((oldId) => { var _a; return (_a = newIdByOldId.get(oldId)) !== null && _a !== void 0 ? _a : oldId; })),
    };
    for (const [oldId, newId] of newIdByOldId.entries()) {
        const oldParentId = nodeIdMapCollection.parentIdById.get(oldId);
        if (oldParentId !== undefined) {
            const newParentId = (_a = newIdByOldId.get(oldParentId)) !== null && _a !== void 0 ? _a : oldParentId;
            collectionDelta.parentIdById.set(newId, {
                oldId,
                value: newParentId,
            });
            if (!collectionDelta.childIdsById.has(newParentId)) {
                collectionDelta.childIdsById.set(newParentId, {
                    oldId: oldParentId,
                    value: updatedChildIds(newIdByOldId, common_1.MapUtils.assertGet(nodeIdMapCollection.childIdsById, oldParentId)),
                });
            }
        }
        const astNode = nodeIdMapCollection.astNodeById.get(oldId);
        let nodeKind;
        if (astNode) {
            collectionDelta.astNodeById.set(newId, astNode);
            nodeKind = astNode.kind;
        }
        else {
            const parseContextNode = common_1.MapUtils.assertGet(nodeIdMapCollection.contextNodeById, oldId, `nodeIdMapCollection has neither astNode nor parseContextNode`);
            collectionDelta.contextNodeById.set(newId, parseContextNode);
            nodeKind = parseContextNode.kind;
        }
        const childIds = nodeIdMapCollection.childIdsById.get(oldId);
        if (childIds && !collectionDelta.childIdsById.has(newId)) {
            for (const childId of childIds) {
                const newChildId = (_b = newIdByOldId.get(childId)) !== null && _b !== void 0 ? _b : childId;
                if (newChildId !== childId) {
                    collectionDelta.parentIdById.set(newChildId, {
                        oldId: childId,
                        value: newChildId,
                    });
                }
            }
            collectionDelta.childIdsById.set(newId, {
                oldId,
                value: updatedChildIds(newIdByOldId, childIds),
            });
        }
        if (!collectionDelta.idsByNodeKind.has(nodeKind)) {
            const oldIdsForNodeKind = common_1.MapUtils.assertGet(nodeIdMapCollection.idsByNodeKind, nodeKind, "expected node kind to exist in idsByNodeKind", {
                oldId,
                nodeKind,
            });
            const newIdsForNodeKind = new Set();
            for (const id of oldIdsForNodeKind) {
                newIdsForNodeKind.add((_c = newIdByOldId.get(id)) !== null && _c !== void 0 ? _c : id);
            }
            collectionDelta.idsByNodeKind.set(nodeKind, newIdsForNodeKind);
        }
    }
    trace.exit();
    return collectionDelta;
}
function applyCollectionDelta(nodeIdMapCollection, collectionDelta, traceManager, correlationId) {
    const trace = traceManager.entry("IdUtils" /* IdUtils */, applyCollectionDelta.name, correlationId);
    for (const [newId, astNode] of collectionDelta.astNodeById.entries()) {
        const oldId = astNode.id;
        const mutableNode = astNode;
        mutableNode.id = newId;
        nodeIdMapCollection.astNodeById.set(newId, astNode);
        if (!collectionDelta.astNodeById.has(oldId)) {
            nodeIdMapCollection.astNodeById.delete(oldId);
        }
    }
    for (const [newId, contextNode] of collectionDelta.contextNodeById.entries()) {
        const oldId = contextNode.id;
        const mutableNode = contextNode;
        mutableNode.id = newId;
        nodeIdMapCollection.contextNodeById.set(newId, contextNode);
        if (!collectionDelta.contextNodeById.has(oldId)) {
            nodeIdMapCollection.contextNodeById.delete(oldId);
        }
    }
    for (const [newId, { oldId, value }] of collectionDelta.parentIdById.entries()) {
        nodeIdMapCollection.parentIdById.set(newId, value);
        if (!collectionDelta.parentIdById.has(oldId)) {
            nodeIdMapCollection.parentIdById.delete(oldId);
        }
    }
    for (const [newId, { oldId, value }] of collectionDelta.childIdsById.entries()) {
        nodeIdMapCollection.childIdsById.set(newId, value);
        if (!collectionDelta.childIdsById.has(oldId)) {
            nodeIdMapCollection.childIdsById.delete(oldId);
        }
    }
    for (const [nodeKind, newIds] of collectionDelta.idsByNodeKind.entries()) {
        nodeIdMapCollection.idsByNodeKind.set(nodeKind, newIds);
    }
    nodeIdMapCollection.leafIds.clear();
    for (const leafId of collectionDelta.leafIds) {
        nodeIdMapCollection.leafIds.add(leafId);
    }
    trace.exit();
}
//# sourceMappingURL=idUtils.js.map