"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.validate = exports.xorNodeTokenRange = exports.hasParsedToken = exports.copy = void 0;
const xorNode_1 = require("../xorNode");
const common_1 = require("../../../common");
const leafSelectors_1 = require("./leafSelectors");
function copy(nodeIdMapCollection) {
    const astNodeById = new Map(Array.from(nodeIdMapCollection.astNodeById.entries()).map(([id, astNode]) => [
        id,
        Object.assign({}, astNode),
    ]));
    const contextNodeById = new Map(Array.from(nodeIdMapCollection.contextNodeById.entries()).map(([id, contextNode]) => [id, Object.assign({}, contextNode)]));
    const rightMostLeaf = nodeIdMapCollection.rightMostLeaf
        ? Object.assign({}, nodeIdMapCollection.rightMostLeaf) : undefined;
    const idsByNodeKind = new Map(Array.from(nodeIdMapCollection.idsByNodeKind.entries()).map(([nodeKind, nodeIds]) => [nodeKind, new Set(nodeIds)]));
    return {
        astNodeById,
        childIdsById: new Map(nodeIdMapCollection.childIdsById),
        contextNodeById,
        leafIds: new Set(nodeIdMapCollection.leafIds),
        idsByNodeKind,
        rightMostLeaf,
        parentIdById: new Map(nodeIdMapCollection.parentIdById),
    };
}
exports.copy = copy;
// Checks if the given nodeId contains at least one parsed token.
function hasParsedToken(nodeIdMapCollection, nodeId) {
    let childIds = nodeIdMapCollection.childIdsById.get(nodeId);
    while (childIds !== undefined) {
        const numChildren = childIds.length;
        // No children means no nothing was parsed under this node.
        if (numChildren === 0) {
            return false;
        }
        // There might be a child under here.
        else if (numChildren === 1) {
            const childId = childIds[0];
            // We know it's an Ast Node, therefore something was parsed.
            if (nodeIdMapCollection.astNodeById.has(childId)) {
                return true;
            }
            // There still might be a child under here. Recurse down to the grandchildren.
            else {
                childIds = nodeIdMapCollection.childIdsById.get(childId);
            }
        }
        // Handles the 'else if (numChildren > 2)' branch.
        // A Context should never have more than one open Context node at a time,
        // meaning there must be at least one Ast node under here.
        else {
            return true;
        }
    }
    return false;
}
exports.hasParsedToken = hasParsedToken;
// Return a XorNodeTokenRange for the given TXorNode.
function xorNodeTokenRange(nodeIdMapCollection, xorNode) {
    return __awaiter(this, void 0, void 0, function* () {
        switch (xorNode.kind) {
            case xorNode_1.XorNodeKind.Ast: {
                const tokenRange = xorNode.node.tokenRange;
                return {
                    tokenIndexStart: tokenRange.tokenIndexStart,
                    tokenIndexEnd: tokenRange.tokenIndexEnd,
                };
            }
            case xorNode_1.XorNodeKind.Context: {
                const contextNode = xorNode.node;
                let tokenIndexEnd;
                const rightMostChild = yield (0, leafSelectors_1.rightMostLeaf)(nodeIdMapCollection, xorNode.node.id);
                if (rightMostChild === undefined) {
                    tokenIndexEnd = contextNode.tokenIndexStart;
                }
                else {
                    tokenIndexEnd = rightMostChild.tokenRange.tokenIndexEnd;
                }
                return {
                    tokenIndexStart: contextNode.tokenIndexStart,
                    tokenIndexEnd,
                };
            }
            default:
                throw common_1.Assert.isNever(xorNode);
        }
    });
}
exports.xorNodeTokenRange = xorNodeTokenRange;
function validate(nodeIdMapCollection) {
    const astNodeById = nodeIdMapCollection.astNodeById;
    const contextNodeById = nodeIdMapCollection.contextNodeById;
    const encounteredNodeKinds = new Set([
        ...Array.from(astNodeById.values()).map((astNode) => astNode.kind),
        ...Array.from(contextNodeById.values()).map((parseContext) => parseContext.kind),
    ]);
    const encounteredIds = new Set([...astNodeById.keys(), ...contextNodeById.keys()]);
    const duplicateIds = [];
    // There's exists some duplicate
    if (encounteredIds.size !== astNodeById.size + contextNodeById.size) {
        for (const nodeId of astNodeById.keys()) {
            if (contextNodeById.has(nodeId)) {
                duplicateIds.push(nodeId);
            }
        }
    }
    const astNodeSummaries = new Map();
    const contextNodeSummaries = new Map();
    for (const [nodeId, astNode] of astNodeById.entries()) {
        astNodeSummaries.set(nodeId, {
            nodeKind: astNode.kind,
            childIds: nodeIdMapCollection.childIdsById.get(nodeId),
            parentId: nodeIdMapCollection.parentIdById.get(nodeId),
            isAstNode: true,
        });
    }
    for (const [nodeId, contextNode] of contextNodeById.entries()) {
        contextNodeSummaries.set(nodeId, {
            nodeKind: contextNode.kind,
            childIds: nodeIdMapCollection.childIdsById.get(nodeId),
            parentId: nodeIdMapCollection.parentIdById.get(nodeId),
            isAstNode: false,
        });
    }
    const nodeIdsByNodeKind = new Map();
    for (const [nodeKind, nodeIds] of nodeIdMapCollection.idsByNodeKind.entries()) {
        nodeIdsByNodeKind.set(nodeKind, Array.from(nodeIds));
    }
    const badParentChildLink = [];
    const unknownParentIdKeys = [];
    const unknownParentIdValues = [];
    for (const [childId, parentId] of nodeIdMapCollection.parentIdById.entries()) {
        if (!encounteredIds.has(parentId)) {
            unknownParentIdKeys.push(parentId);
        }
        if (!encounteredIds.has(childId)) {
            unknownParentIdValues.push(childId);
        }
        const childIdsOfParent = nodeIdMapCollection.childIdsById.get(parentId);
        if (childIdsOfParent !== undefined && !childIdsOfParent.includes(childId)) {
            badParentChildLink.push([parentId, childId]);
        }
    }
    const unknownChildIdsKeys = [];
    const unknownChildIdsValues = [];
    for (const [parentId, childIds] of nodeIdMapCollection.childIdsById.entries()) {
        if (!encounteredIds.has(parentId)) {
            unknownChildIdsKeys.push(parentId);
        }
        for (const childId of childIds) {
            if (!encounteredIds.has(childId)) {
                unknownChildIdsValues.push(childId);
            }
            if (nodeIdMapCollection.parentIdById.get(childId) !== parentId) {
                badParentChildLink.push([parentId, childId]);
            }
        }
    }
    const unknownByNodeKindNodeIds = [];
    for (const nodeIds of nodeIdMapCollection.idsByNodeKind.values()) {
        for (const nodeId of nodeIds) {
            if (!encounteredIds.has(nodeId)) {
                unknownByNodeKindNodeIds.push(nodeId);
            }
        }
    }
    return {
        astNodes: astNodeSummaries,
        contextNodes: contextNodeSummaries,
        leafIds: Array.from(nodeIdMapCollection.leafIds),
        nodeIdsByNodeKind,
        badParentChildLink,
        duplicateIds,
        unknownByNodeKindNodeIds,
        unknownByNodeKindNodeKinds: Array.from(nodeIdMapCollection.idsByNodeKind.keys()).filter((nodeKind) => !encounteredNodeKinds.has(nodeKind)),
        unknownChildIdsKeys,
        unknownChildIdsValues,
        unknownLeafIds: Array.from(nodeIdMapCollection.leafIds).filter((id) => !encounteredIds.has(id)),
        unknownParentIdKeys,
        unknownParentIdValues,
    };
}
exports.validate = validate;
//# sourceMappingURL=nodeIdMapUtils.js.map