"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.parameterIdentifierLiteral = exports.parameterIdentifier = exports.invokeExpressionIdentifierLiteral = exports.invokeExpressionIdentifier = exports.assertRecursiveExpressionPreviousSibling = void 0;
const childSelectors_1 = require("./childSelectors");
const parentSelectors_1 = require("./parentSelectors");
const commonSelectors_1 = require("./commonSelectors");
const __1 = require("..");
const language_1 = require("../../../language");
const common_1 = require("../../../common");
// Returns the previous sibling of the given recursive expression.
// Commonly used for things like getting the identifier name used in an InvokeExpression.
// Returns undefined if there is no previous sibling.
function assertRecursiveExpressionPreviousSibling(nodeIdMapCollection, nodeId, expectedNodeKinds) {
    const xorNode = (0, commonSelectors_1.assertXor)(nodeIdMapCollection, nodeId);
    const arrayWrapper = (0, parentSelectors_1.assertParentXorChecked)(nodeIdMapCollection, nodeId, language_1.Ast.NodeKind.ArrayWrapper);
    const primaryExpressionAttributeId = xorNode.node.attributeIndex;
    // It's not the first element in the ArrayWrapper.
    if (primaryExpressionAttributeId && primaryExpressionAttributeId > 0) {
        const childIds = __1.NodeIdMapUtils.assertChildIds(nodeIdMapCollection.childIdsById, arrayWrapper.node.id);
        const indexOfPrimaryExpressionId = childIds.indexOf(xorNode.node.id);
        if (indexOfPrimaryExpressionId === -1 || indexOfPrimaryExpressionId === 0) {
            const details = {
                xorNodeId: xorNode.node.id,
                arrayWrapperId: arrayWrapper.node.id,
                indexOfPrimaryExpressionId,
            };
            throw new common_1.CommonError.InvariantError(`expected to find xorNodeId in arrayWrapper's children at an index > 0`, details);
        }
        return expectedNodeKinds
            ? (0, childSelectors_1.assertNthChildXorChecked)(nodeIdMapCollection, arrayWrapper.node.id, indexOfPrimaryExpressionId - 1, expectedNodeKinds)
            : (0, childSelectors_1.assertNthChildXor)(nodeIdMapCollection, arrayWrapper.node.id, indexOfPrimaryExpressionId - 1);
    }
    // It's the first element in ArrayWrapper, meaning we must grab RecursivePrimaryExpression.head
    else {
        const recursivePrimaryExpression = (0, parentSelectors_1.assertParentXor)(nodeIdMapCollection, arrayWrapper.node.id);
        return expectedNodeKinds
            ? (0, childSelectors_1.assertNthChildXorChecked)(nodeIdMapCollection, recursivePrimaryExpression.node.id, 0, expectedNodeKinds)
            : (0, childSelectors_1.assertNthChildXor)(nodeIdMapCollection, recursivePrimaryExpression.node.id, 0);
    }
}
exports.assertRecursiveExpressionPreviousSibling = assertRecursiveExpressionPreviousSibling;
// Asserts the given node is an InvokeExpression,
// then returns the previous sibling on the condition that it's an identifier.
function invokeExpressionIdentifier(nodeIdMapCollection, nodeId) {
    const invokeExprXorNode = (0, commonSelectors_1.assertXorChecked)(nodeIdMapCollection, nodeId, language_1.Ast.NodeKind.InvokeExpression);
    // The only place for an identifier in a RecursivePrimaryExpression is as the head, therefore an InvokeExpression
    // only has a name if the InvokeExpression is the 0th element in the RecursivePrimaryExpressionArray.
    if (invokeExprXorNode.node.attributeIndex !== 0) {
        return undefined;
    }
    // Grab the RecursivePrimaryExpression's head if it's an IdentifierExpression
    const recursiveArrayXorNode = (0, parentSelectors_1.assertParentXor)(nodeIdMapCollection, invokeExprXorNode.node.id);
    const recursiveExprXorNode = (0, parentSelectors_1.assertParentXor)(nodeIdMapCollection, recursiveArrayXorNode.node.id);
    const headXorNode = (0, childSelectors_1.nthChildXorChecked)(nodeIdMapCollection, recursiveExprXorNode.node.id, 0, language_1.Ast.NodeKind.IdentifierExpression);
    // It's not an identifier expression so there's nothing we can do.
    if (headXorNode === undefined) {
        return undefined;
    }
    // The only place for an identifier in a RecursivePrimaryExpression is as the head, therefore an InvokeExpression
    // only has a name if the InvokeExpression is the 0th element in the RecursivePrimaryExpressionArray.
    if (__1.XorNodeUtils.isContext(headXorNode)) {
        const details = {
            identifierExpressionNodeId: headXorNode.node.id,
            invokeExpressionNodeId: invokeExprXorNode.node.id,
        };
        throw new common_1.CommonError.InvariantError(`the younger IdentifierExpression sibling should've finished parsing before the InvokeExpression node was reached`, details);
    }
    return headXorNode;
}
exports.invokeExpressionIdentifier = invokeExpressionIdentifier;
// Unboxes the identifier literal for function name if it exists.
function invokeExpressionIdentifierLiteral(nodeIdMapCollection, nodeId) {
    (0, commonSelectors_1.assertXorChecked)(nodeIdMapCollection, nodeId, language_1.Ast.NodeKind.InvokeExpression);
    const identifierExpressionXorNode = invokeExpressionIdentifier(nodeIdMapCollection, nodeId);
    if (identifierExpressionXorNode === undefined || __1.XorNodeUtils.isContext(identifierExpressionXorNode)) {
        return undefined;
    }
    const identifierExpression = identifierExpressionXorNode.node;
    return identifierExpression.inclusiveConstant === undefined
        ? identifierExpression.identifier.literal
        : identifierExpression.inclusiveConstant.constantKind + identifierExpression.identifier.literal;
}
exports.invokeExpressionIdentifierLiteral = invokeExpressionIdentifierLiteral;
// Unboxes the node if it's a identifier
function parameterIdentifier(nodeIdMapCollection, functionParameter) {
    return (0, childSelectors_1.nthChildAstChecked)(nodeIdMapCollection, functionParameter.node.id, 1, language_1.Ast.NodeKind.Identifier);
}
exports.parameterIdentifier = parameterIdentifier;
// Unboxes the identifier literal if it exists.
function parameterIdentifierLiteral(nodeIdMapCollection, functionParameter) {
    const identifier = parameterIdentifier(nodeIdMapCollection, functionParameter);
    return identifier ? identifier.literal : undefined;
}
exports.parameterIdentifierLiteral = parameterIdentifierLiteral;
//# sourceMappingURL=specializedSelectors.js.map