"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.restoreCheckpoint = exports.checkpoint = exports.tryParseDocument = exports.tryParse = void 0;
const common_1 = require("../../common");
const nodeIdMap_1 = require("../nodeIdMap");
const context_1 = require("../context");
const __1 = require("..");
const parseState_1 = require("../parseState");
function tryParse(parseSettings, lexerSnapshot) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = parseSettings.traceManager.entry("ParserUtils" /* ParserUtils */, tryParse.name, parseSettings.initialCorrelationId);
        const updatedSettings = Object.assign(Object.assign({}, parseSettings), { initialCorrelationId: trace.id });
        const parserEntryPoint = updatedSettings === null || updatedSettings === void 0 ? void 0 : updatedSettings.parserEntryPoint;
        if (parserEntryPoint === undefined) {
            return yield tryParseDocument(updatedSettings, lexerSnapshot);
        }
        const parseState = updatedSettings.newParseState(lexerSnapshot, defaultOverrides(updatedSettings));
        try {
            const root = yield parserEntryPoint(parseState, updatedSettings.parser, trace.id);
            parseState_1.ParseStateUtils.assertIsDoneParsing(parseState);
            return common_1.ResultUtils.ok({
                lexerSnapshot,
                root,
                state: parseState,
            });
        }
        catch (caught) {
            common_1.Assert.isInstanceofError(caught);
            common_1.CommonError.throwIfCancellationError(caught);
            return common_1.ResultUtils.error(ensureParseError(parseState, caught, updatedSettings.locale));
        }
    });
}
exports.tryParse = tryParse;
// Attempts to parse the document both as an expression and section document.
// Whichever attempt consumed the most tokens is the one returned. Ties go to expression documents.
function tryParseDocument(parseSettings, lexerSnapshot) {
    return __awaiter(this, void 0, void 0, function* () {
        const trace = parseSettings.traceManager.entry("ParserUtils" /* ParserUtils */, tryParseDocument.name, parseSettings.initialCorrelationId);
        let root;
        const expressionDocumentState = parseSettings.newParseState(lexerSnapshot, defaultOverrides(parseSettings));
        try {
            root = yield parseSettings.parser.readExpression(expressionDocumentState, parseSettings.parser, trace.id);
            parseState_1.ParseStateUtils.assertIsDoneParsing(expressionDocumentState);
            trace.exit();
            return common_1.ResultUtils.ok({
                lexerSnapshot,
                root,
                state: expressionDocumentState,
            });
        }
        catch (expressionDocumentError) {
            common_1.Assert.isInstanceofError(expressionDocumentError);
            common_1.CommonError.throwIfCancellationError(expressionDocumentError);
            const sectionDocumentState = parseSettings.newParseState(lexerSnapshot, defaultOverrides(parseSettings));
            try {
                root = yield parseSettings.parser.readSectionDocument(sectionDocumentState, parseSettings.parser, trace.id);
                parseState_1.ParseStateUtils.assertIsDoneParsing(sectionDocumentState);
                trace.exit();
                return common_1.ResultUtils.ok({
                    lexerSnapshot,
                    root,
                    state: sectionDocumentState,
                });
            }
            catch (sectionDocumentError) {
                common_1.Assert.isInstanceofError(sectionDocumentError);
                common_1.CommonError.throwIfCancellationError(expressionDocumentError);
                let betterParsedState;
                let betterParsedError;
                if (expressionDocumentState.tokenIndex >= sectionDocumentState.tokenIndex) {
                    betterParsedState = expressionDocumentState;
                    betterParsedError = expressionDocumentError;
                }
                else {
                    betterParsedState = sectionDocumentState;
                    betterParsedError = sectionDocumentError;
                }
                trace.exit();
                return common_1.ResultUtils.error(ensureParseError(betterParsedState, betterParsedError, parseSettings.locale));
            }
        }
    });
}
exports.tryParseDocument = tryParseDocument;
// If you have a custom parser + parser state,
// then you'll have to create your own checkpoint/restoreCheckpoint functions.
//
// Due to performance reasons the backup no longer can include a naive deep copy of the context state.
// Instead it's assumed that a backup is made immediately before a try/catch read block.
// This means the state begins in a parsing context and the backup will either be immediately consumed or dropped.
// Therefore we only care about the delta between before and after the try/catch block.
// Thanks to the invariants above and the fact the ids for nodes are an auto-incrementing integer
// we can easily delete all context nodes past the id of when the backup was created.
// eslint-disable-next-line require-await
function checkpoint(state) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        return {
            tokenIndex: state.tokenIndex,
            contextStateIdCounter: state.contextState.idCounter,
            contextNodeId: (_a = state.currentContextNode) === null || _a === void 0 ? void 0 : _a.id,
        };
    });
}
exports.checkpoint = checkpoint;
// If you have a custom parser + parser state,
// then you'll have to create your own (create|restore)Checkpoint functions.
// See createCheckpoint above for more information.
// eslint-disable-next-line require-await
function restoreCheckpoint(state, checkpoint) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        state.tokenIndex = checkpoint.tokenIndex;
        state.currentToken = state.lexerSnapshot.tokens[state.tokenIndex];
        state.currentTokenKind = (_a = state.currentToken) === null || _a === void 0 ? void 0 : _a.kind;
        const contextState = state.contextState;
        const nodeIdMapCollection = state.contextState.nodeIdMapCollection;
        const backupIdCounter = checkpoint.contextStateIdCounter;
        contextState.idCounter = backupIdCounter;
        const newContextNodeIds = [];
        const newAstNodeIds = [];
        for (const nodeId of nodeIdMapCollection.astNodeById.keys()) {
            if (nodeId > backupIdCounter) {
                newAstNodeIds.push(nodeId);
            }
        }
        for (const nodeId of nodeIdMapCollection.contextNodeById.keys()) {
            if (nodeId > backupIdCounter) {
                newContextNodeIds.push(nodeId);
            }
        }
        const reverseNumberSort = (left, right) => right - left;
        for (const nodeId of newAstNodeIds.sort(reverseNumberSort)) {
            const parentId = nodeIdMapCollection.parentIdById.get(nodeId);
            const parentWillBeDeleted = parentId !== undefined && parentId >= backupIdCounter;
            context_1.ParseContextUtils.deleteAst(state.contextState, nodeId, parentWillBeDeleted);
        }
        for (const nodeId of newContextNodeIds.sort(reverseNumberSort)) {
            context_1.ParseContextUtils.deleteContext(state.contextState, nodeId);
        }
        if (checkpoint.contextNodeId) {
            state.currentContextNode = nodeIdMap_1.NodeIdMapUtils.assertContext(state.contextState.nodeIdMapCollection, checkpoint.contextNodeId);
        }
        else {
            state.currentContextNode = undefined;
        }
    });
}
exports.restoreCheckpoint = restoreCheckpoint;
function ensureParseError(state, error, locale) {
    if (error instanceof __1.ParseError.ParseError) {
        return error;
    }
    else if (__1.ParseError.isTInnerParseError(error)) {
        return new __1.ParseError.ParseError(error, state);
    }
    else {
        return common_1.CommonError.ensureCommonError(error, locale);
    }
}
function defaultOverrides(parseSettings) {
    return {
        locale: parseSettings.locale,
        cancellationToken: parseSettings.cancellationToken,
        traceManager: parseSettings.traceManager,
    };
}
//# sourceMappingURL=parserUtils.js.map