"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
Object.defineProperty(exports, "__esModule", { value: true });
exports.combineOperatorsAndOperands = void 0;
const common_1 = require("../../../common");
const language_1 = require("../../../language");
const commonTypes_1 = require("./commonTypes");
const __1 = require("../..");
const __2 = require("..");
// Takes N operators and N+1 operands.
// Continually combines the highest precedence operators the operands they're adjacent to,
// thus reducing the number of operators and operands by 1 per iteration.
// Ends with 0 operators and 1 operand.
function combineOperatorsAndOperands(state, parser, placeholderContextNodeId, operatorsAndOperands, correlationId) {
    let { operatorConstants, operands } = operatorsAndOperands;
    common_1.Assert.isTrue(operatorConstants.length + 1 === operands.length, `operators.length !== (operands.length + 1) failed`, {
        operandsLength: operands.length,
        operatorsLength: operatorConstants.length,
    });
    const trace = state.traceManager.entry(commonTypes_1.CombinatorialParserV2TraceConstant.CombinatorialParseV2, combineOperatorsAndOperands.name, correlationId);
    // Sorts operator by precedence where the highest precedence operators are first.
    const sortedOperatorConstants = sortByPrecedence(operatorConstants);
    const nodeIdMapCollection = state.contextState.nodeIdMapCollection;
    const numOperators = sortedOperatorConstants.length;
    for (const [precedenceSortableOperatorConstant, index] of common_1.ArrayUtils.enumerate(sortedOperatorConstants)) {
        const { leftOperandIndex, operatorConstant } = precedenceSortableOperatorConstant;
        const nodeKind = language_1.AstUtils.nodeKindFromTBinOpExpressionOperator(operatorConstant.constantKind);
        const left = common_1.ArrayUtils.assertGet(operands, leftOperandIndex);
        const right = common_1.ArrayUtils.assertGet(operands, leftOperandIndex + 1);
        const binOpParseContext = __1.ParseContextUtils.startContext(state.contextState, nodeKind, left.tokenRange.tokenIndexStart, state.lexerSnapshot.tokens[left.tokenRange.tokenIndexStart], undefined);
        placeParseContextUnderPlaceholderContext(state, binOpParseContext, placeholderContextNodeId);
        const validator = validatorForOperator(operatorConstant.constantKind);
        if (!validator.validateLeftOperand(left)) {
            setParseStateToNodeStart(state, left);
            validator.fallbackLeftOperand(state, parser, trace.id);
            throw new common_1.CommonError.InvariantError(`validateLeftOperand failed and then fallbackLeftOperand did not throw.`);
        }
        addAstAsChild(nodeIdMapCollection, binOpParseContext, left);
        addAstAsChild(nodeIdMapCollection, binOpParseContext, operatorConstant);
        if (!validator.validateRightOperand(right)) {
            setParseStateToAfterNodeEnd(state, operatorConstant);
            validator.fallbackRightOperand(state, parser, trace.id);
            throw new common_1.CommonError.InvariantError(`validateRightOperand failed and then fallbackRightOperand did not throw.`);
        }
        addAstAsChild(nodeIdMapCollection, binOpParseContext, right);
        setParseStateToAfterNodeEnd(state, right);
        const leftTokenRange = left.tokenRange;
        const rightTokenRange = right.tokenRange;
        // We started with an operatorConstant belonging under TBinOpExpression["operatorConstant"].
        // Working backwards, we derived the nodeKind for some TBinOpExpression `T` and validators
        // which should validate left is `T["left"]` and right is `T["right"]`.
        // Therefore, we should be able to cast it as a TBinOpExpression.
        const binOp = {
            kind: nodeKind,
            id: binOpParseContext.id,
            attributeIndex: 0,
            tokenRange: {
                tokenIndexStart: leftTokenRange.tokenIndexStart,
                tokenIndexEnd: rightTokenRange.tokenIndexEnd,
                positionStart: leftTokenRange.positionStart,
                positionEnd: rightTokenRange.positionEnd,
            },
            isLeaf: false,
            left,
            operatorConstant,
            right,
        };
        // Promote from ParseContext into Ast
        nodeIdMapCollection.astNodeById.set(binOp.id, binOp);
        common_1.MapUtils.assertDelete(nodeIdMapCollection.contextNodeById, binOp.id);
        // Modify the operands and operatorConstants for the next iteration by:
        //  - replacing the `left` and `right` operands with the new `binOp` node
        //  - removing the operator
        operands = [...operands.slice(0, leftOperandIndex), binOp, ...operands.slice(leftOperandIndex + 2)];
        operatorConstants = common_1.ArrayUtils.assertRemoveAtIndex(operatorConstants, leftOperandIndex);
        // Since we've mutated the list of operands we need to update the leftOperandIndex.
        // for all operators to the right of the one we just processed.
        for (let unvisitedIndex = index + 1; unvisitedIndex < numOperators; unvisitedIndex += 1) {
            const unvisitedOperator = common_1.ArrayUtils.assertGet(sortedOperatorConstants, unvisitedIndex);
            if (unvisitedOperator.leftOperandIndex > leftOperandIndex) {
                unvisitedOperator.leftOperandIndex -= 1;
            }
        }
    }
    const result = common_1.ArrayUtils.assertGet(operands, 0);
    common_1.Assert.isTrue(language_1.AstUtils.isTBinOpExpression(result) || language_1.AstUtils.isTUnaryExpression(result), `AstUtils.isTBinOpExpression(result) || AstUtils.isTUnaryExpression(result) failed`, { resultNodeKind: result.kind });
    trace.exit();
    return result;
}
exports.combineOperatorsAndOperands = combineOperatorsAndOperands;
const ValidatorForAsExpression = {
    tag: "AsExpression",
    validateLeftOperand: (node) => language_1.AstUtils.isTAsExpression(node),
    validateRightOperand: (node) => language_1.AstUtils.isTNullablePrimitiveType(node),
    fallbackLeftOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readEqualityExpression(state, parser, correlationId),
    fallbackRightOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readNullablePrimitiveType(state, parser, correlationId),
};
const ValidatorForEqualityExpressionAndBelow = {
    tag: "EqualityExpression",
    validateLeftOperand: (node) => language_1.AstUtils.isTEqualityExpression(node),
    validateRightOperand: (node) => language_1.AstUtils.isTEqualityExpression(node),
    fallbackLeftOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readMetadataExpression(state, parser, correlationId),
    fallbackRightOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readMetadataExpression(state, parser, correlationId),
};
const ValidatorForIsExpression = {
    tag: "IsExpression",
    validateLeftOperand: (node) => language_1.AstUtils.isTIsExpression(node),
    validateRightOperand: (node) => language_1.AstUtils.isTNullablePrimitiveType(node),
    fallbackLeftOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readLogicalExpression(state, parser, correlationId),
    fallbackRightOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readNullablePrimitiveType(state, parser, correlationId),
};
const ValidatorForLogicalAndExpression = {
    tag: "LogicalAndExpression",
    validateLeftOperand: (node) => language_1.AstUtils.isTLogicalExpression(node),
    validateRightOperand: (node) => language_1.AstUtils.isTIsExpression(node),
    fallbackLeftOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readIsExpression(state, parser, correlationId),
    fallbackRightOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readIsExpression(state, parser, correlationId),
};
const ValidatorForLogicalOrExpression = {
    tag: "LogicalOrExpression",
    validateLeftOperand: (node) => language_1.AstUtils.isTLogicalExpression(node),
    validateRightOperand: isTIsExpressionOrLogicalAndExpression,
    fallbackLeftOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readLogicalExpression(state, parser, correlationId),
    fallbackRightOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readLogicalExpression(state, parser, correlationId),
};
const ValidatorForNullCoalescingExpression = {
    tag: "NullCoalescingExpression",
    validateLeftOperand: (node) => language_1.AstUtils.isTLogicalExpression(node),
    validateRightOperand: (node) => language_1.AstUtils.isTNullCoalescingExpression(node),
    fallbackLeftOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readLogicalExpression(state, parser, correlationId),
    fallbackRightOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readLogicalExpression(state, parser, correlationId),
};
const ValidatorForMetadataExpression = {
    tag: "MetadataExpression",
    validateLeftOperand: (node) => language_1.AstUtils.isTUnaryExpression(node),
    validateRightOperand: (node) => language_1.AstUtils.isTUnaryExpression(node),
    fallbackLeftOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readUnaryExpression(state, parser, correlationId),
    fallbackRightOperand: (state, parser, correlationId) => __2.NaiveParseSteps.readUnaryExpression(state, parser, correlationId),
};
function addAstAsChild(nodeIdMapCollection, parent, child) {
    parent.attributeCounter += 1;
    const parentId = parent.id;
    const oldChildren = nodeIdMapCollection.childIdsById.get(parentId);
    nodeIdMapCollection.astNodeById.set(child.id, child);
    nodeIdMapCollection.parentIdById.set(child.id, parentId);
    nodeIdMapCollection.childIdsById.set(parentId, [...(oldChildren !== null && oldChildren !== void 0 ? oldChildren : []), child.id]);
    if (child.isLeaf) {
        nodeIdMapCollection.leafIds.add(child.id);
    }
    addNodeKindToCollection(nodeIdMapCollection.idsByNodeKind, child.kind, child.id);
}
function addNodeKindToCollection(idsByNodeKind, nodeKind, nodeId) {
    const collection = idsByNodeKind.get(nodeKind);
    if (collection) {
        collection.add(nodeId);
    }
    else {
        idsByNodeKind.set(nodeKind, new Set([nodeId]));
    }
}
function isTIsExpressionOrLogicalAndExpression(node) {
    return (language_1.AstUtils.isTIsExpression(node) ||
        (language_1.AstUtils.isNodeKind(node, language_1.Ast.NodeKind.LogicalExpression) &&
            node.operatorConstant.constantKind === language_1.Constant.LogicalOperator.And));
}
function placeParseContextUnderPlaceholderContext(state, parseContext, placeholderContextNodeId) {
    state.currentContextNode = parseContext;
    const nodeIdMapCollection = state.contextState.nodeIdMapCollection;
    nodeIdMapCollection.childIdsById.set(placeholderContextNodeId, [parseContext.id]);
    nodeIdMapCollection.parentIdById.set(parseContext.id, placeholderContextNodeId);
}
function setParseStateToNodeStart(state, node) {
    setParseStateToTokenIndex(state, node.tokenRange.tokenIndexStart);
}
function setParseStateToAfterNodeEnd(state, node) {
    setParseStateToTokenIndex(state, node.tokenRange.tokenIndexEnd + 1);
}
function setParseStateToTokenIndex(state, tokenIndex) {
    const token = state.lexerSnapshot.tokens[tokenIndex];
    state.currentToken = token;
    state.currentTokenKind = token === null || token === void 0 ? void 0 : token.kind;
    state.tokenIndex = tokenIndex;
}
// Returns an array that iterates over the operators in order of precedence,
// with tie breakers being the order they were given (which should also be the order they appear in the document)
function sortByPrecedence(operators) {
    const sortableOperatorConstant = operators.map((operatorConstant, index) => ({
        leftOperandIndex: index,
        operatorConstant,
        precedence: language_1.ConstantUtils.binOpExpressionOperatorPrecedence(operatorConstant.constantKind),
    }));
    return sortableOperatorConstant.sort((left, right) => {
        const precedenceDelta = right.precedence - left.precedence;
        return precedenceDelta === 0 ? left.leftOperandIndex - right.leftOperandIndex : precedenceDelta;
    });
}
function validatorForOperator(operator) {
    switch (operator) {
        case language_1.Constant.ArithmeticOperator.Multiplication:
        case language_1.Constant.ArithmeticOperator.Division:
        case language_1.Constant.ArithmeticOperator.Addition:
        case language_1.Constant.ArithmeticOperator.Subtraction:
        case language_1.Constant.ArithmeticOperator.And:
        case language_1.Constant.RelationalOperator.LessThan:
        case language_1.Constant.RelationalOperator.LessThanEqualTo:
        case language_1.Constant.RelationalOperator.GreaterThan:
        case language_1.Constant.RelationalOperator.GreaterThanEqualTo:
        case language_1.Constant.EqualityOperator.EqualTo:
        case language_1.Constant.EqualityOperator.NotEqualTo:
            return ValidatorForEqualityExpressionAndBelow;
        case language_1.Constant.KeywordConstant.Meta:
            return ValidatorForMetadataExpression;
        case language_1.Constant.KeywordConstant.As:
            return ValidatorForAsExpression;
        case language_1.Constant.KeywordConstant.Is:
            return ValidatorForIsExpression;
        case language_1.Constant.LogicalOperator.And:
            return ValidatorForLogicalAndExpression;
        case language_1.Constant.LogicalOperator.Or:
            return ValidatorForLogicalOrExpression;
        case language_1.Constant.MiscConstant.NullCoalescingOperator:
            return ValidatorForNullCoalescingExpression;
        default:
            throw common_1.Assert.isNever(operator);
    }
}
//# sourceMappingURL=combineOperatorsAndOperands.js.map