"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.readOperatorsAndOperands = void 0;
const common_1 = require("../../../common");
const language_1 = require("../../../language");
const commonTypes_1 = require("./commonTypes");
const parseState_1 = require("../../parseState");
const __1 = require("..");
// Asserts that N operators and N+1 operands are read (where N >= 0).
// Both operators and operands are read in a left-to-right fashion and are returned in that order.
//
// It's possible that the reading of the operand to fail. In that case you'll see the following:
//  - a binary expression context under the initial context node
//  - it will have 2 children
//      - an Ast node for the operatorConstant
//      - a parse context node for the operand
function readOperatorsAndOperands(state, parser, correlationId) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const trace = state.traceManager.entry(commonTypes_1.CombinatorialParserV2TraceConstant.CombinatorialParseV2, readOperatorsAndOperands.name, correlationId);
        const initialCurrentContextNode = common_1.Assert.asDefined(state.currentContextNode);
        const nodeIdMapCollection = state.contextState.nodeIdMapCollection;
        const operatorConstants = [];
        // The initial read will always be a TUnaryExpression.
        // Continually read an operator and operand while a valid operator constant exists.
        // A different operand reader is used depending on the operator constant.
        const operands = [
            yield parser.readUnaryExpression(state, parser, trace.id),
        ];
        let nextRead = state.currentTokenKind
            ? nextOperatorAndOperandRead(state.currentTokenKind)
            : undefined;
        while (nextRead) {
            const iterativeParseContext = parseState_1.ParseStateUtils.startContext(state, nextRead.binOpExpressionNodeKind);
            iterativeParseContext.attributeCounter = 1;
            const operatorConstant = __1.NaiveParseSteps.readTokenKindAsConstant(state, nextRead.operatorTokenKind, nextRead.operatorConstantKind, trace.id);
            let operand;
            switch (nextRead.operandNodeKind) {
                case language_1.Ast.NodeKind.UnaryExpression: {
                    // eslint-disable-next-line no-await-in-loop
                    operand = yield parser.readUnaryExpression(state, parser, trace.id);
                    break;
                }
                case language_1.Ast.NodeKind.NullablePrimitiveType:
                    // eslint-disable-next-line no-await-in-loop
                    operand = yield parser.readNullablePrimitiveType(state, parser, trace.id);
                    break;
                case language_1.Ast.NodeKind.LogicalExpression:
                    // eslint-disable-next-line no-await-in-loop
                    operand = yield parser.readLogicalExpression(state, parser, trace.id);
                    break;
                default:
                    throw common_1.Assert.isNever(nextRead.operandNodeKind);
            }
            // Append the operator/operator for the result.
            operatorConstants.push(operatorConstant);
            operands.push(operand);
            // If we don't do any nodeIdMapCollection modifications it leave a weird state where
            // every operator/operand would be a child of the initial context node,
            // which doesn't make much sense if you had to examine the parse state downstream.
            //
            // For now we're deleting those links and will re-create them later as we combine operators/operands.
            for (const nodeId of [operand.id, operatorConstant.id, iterativeParseContext.id]) {
                nodeIdMapCollection.astNodeById.delete(nodeId);
                nodeIdMapCollection.parentIdById.delete(nodeId);
            }
            nodeIdMapCollection.childIdsById.set(initialCurrentContextNode.id, common_1.ArrayUtils.assertRemoveFirstInstance(common_1.MapUtils.assertGet(nodeIdMapCollection.childIdsById, initialCurrentContextNode.id), iterativeParseContext.id));
            nodeIdMapCollection.contextNodeById.delete(iterativeParseContext.id);
            nodeIdMapCollection.childIdsById.delete(iterativeParseContext.id);
            nodeIdMapCollection.leafIds.delete(operatorConstant.id);
            nodeIdMapCollection.leafIds.delete(operand.id);
            removeIdFromIdsByNodeKind(nodeIdMapCollection.idsByNodeKind, iterativeParseContext.kind, iterativeParseContext.id);
            removeIdFromIdsByNodeKind(nodeIdMapCollection.idsByNodeKind, language_1.Ast.NodeKind.Constant, operatorConstant.id);
            removeIdFromIdsByNodeKind(nodeIdMapCollection.idsByNodeKind, operand.kind, operand.id);
            // eslint-disable-next-line require-atomic-updates
            state.currentContextNode = initialCurrentContextNode;
            nextRead = state.currentTokenKind ? nextOperatorAndOperandRead(state.currentTokenKind) : undefined;
        }
        common_1.Assert.isTrue(((_a = state.currentContextNode) === null || _a === void 0 ? void 0 : _a.id) === initialCurrentContextNode.id, `state.currentContextNode.id === initialCurrentContextNode failed`, {
            currentContextNodeId: (_b = state.currentContextNode) === null || _b === void 0 ? void 0 : _b.id,
            initialCurrentContextNodeId: initialCurrentContextNode.id,
        });
        common_1.Assert.isTrue(operatorConstants.length + 1 === operands.length, `operatorConstants.length + 1 === operands.length failed`, {
            operatorsLength: operatorConstants.length,
            operandsLength: operands.length,
        });
        trace.exit();
        return {
            operatorConstants,
            operands,
        };
    });
}
exports.readOperatorsAndOperands = readOperatorsAndOperands;
function nextOperatorAndOperandRead(tokenKind) {
    switch (tokenKind) {
        case language_1.Token.TokenKind.Asterisk:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.ArithmeticExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.ArithmeticOperator.Multiplication,
                operatorTokenKind: language_1.Token.TokenKind.Asterisk,
            };
        case language_1.Token.TokenKind.Division:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.ArithmeticExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.ArithmeticOperator.Division,
                operatorTokenKind: language_1.Token.TokenKind.Division,
            };
        case language_1.Token.TokenKind.Plus:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.ArithmeticExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.ArithmeticOperator.Addition,
                operatorTokenKind: language_1.Token.TokenKind.Plus,
            };
        case language_1.Token.TokenKind.Minus:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.ArithmeticExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.ArithmeticOperator.Subtraction,
                operatorTokenKind: language_1.Token.TokenKind.Minus,
            };
        case language_1.Token.TokenKind.Ampersand:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.ArithmeticExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.ArithmeticOperator.And,
                operatorTokenKind: language_1.Token.TokenKind.Ampersand,
            };
        case language_1.Token.TokenKind.Equal:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.EqualityExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.EqualityOperator.EqualTo,
                operatorTokenKind: language_1.Token.TokenKind.Equal,
            };
        case language_1.Token.TokenKind.NotEqual:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.EqualityExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.EqualityOperator.NotEqualTo,
                operatorTokenKind: language_1.Token.TokenKind.NotEqual,
            };
        case language_1.Token.TokenKind.KeywordAnd:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.LogicalExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.LogicalOperator.And,
                operatorTokenKind: language_1.Token.TokenKind.KeywordAnd,
            };
        case language_1.Token.TokenKind.KeywordOr:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.LogicalExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.LogicalOperator.Or,
                operatorTokenKind: language_1.Token.TokenKind.KeywordOr,
            };
        case language_1.Token.TokenKind.LessThan:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.RelationalExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.RelationalOperator.LessThan,
                operatorTokenKind: language_1.Token.TokenKind.LessThan,
            };
        case language_1.Token.TokenKind.LessThanEqualTo:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.RelationalExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.RelationalOperator.LessThanEqualTo,
                operatorTokenKind: language_1.Token.TokenKind.LessThanEqualTo,
            };
        case language_1.Token.TokenKind.GreaterThan:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.RelationalExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.RelationalOperator.GreaterThan,
                operatorTokenKind: language_1.Token.TokenKind.GreaterThan,
            };
        case language_1.Token.TokenKind.GreaterThanEqualTo:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.RelationalExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.RelationalOperator.GreaterThanEqualTo,
                operatorTokenKind: language_1.Token.TokenKind.GreaterThanEqualTo,
            };
        case language_1.Token.TokenKind.KeywordAs:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.AsExpression,
                operandNodeKind: language_1.Ast.NodeKind.NullablePrimitiveType,
                operatorConstantKind: language_1.Constant.KeywordConstant.As,
                operatorTokenKind: language_1.Token.TokenKind.KeywordAs,
            };
        case language_1.Token.TokenKind.KeywordIs:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.IsExpression,
                operandNodeKind: language_1.Ast.NodeKind.NullablePrimitiveType,
                operatorConstantKind: language_1.Constant.KeywordConstant.Is,
                operatorTokenKind: language_1.Token.TokenKind.KeywordIs,
            };
        case language_1.Token.TokenKind.KeywordMeta:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.MetadataExpression,
                operandNodeKind: language_1.Ast.NodeKind.UnaryExpression,
                operatorConstantKind: language_1.Constant.KeywordConstant.Meta,
                operatorTokenKind: language_1.Token.TokenKind.KeywordMeta,
            };
        case language_1.Token.TokenKind.NullCoalescingOperator:
            return {
                binOpExpressionNodeKind: language_1.Ast.NodeKind.NullCoalescingExpression,
                operandNodeKind: language_1.Ast.NodeKind.LogicalExpression,
                operatorConstantKind: language_1.Constant.MiscConstant.NullCoalescingOperator,
                operatorTokenKind: language_1.Token.TokenKind.NullCoalescingOperator,
            };
        default:
            return undefined;
    }
}
function removeIdFromIdsByNodeKind(idsByNodeKind, nodeKind, nodeId) {
    const collection = common_1.MapUtils.assertGet(idsByNodeKind, nodeKind);
    common_1.SetUtils.assertDelete(collection, nodeId);
    if (collection.size === 0) {
        idsByNodeKind.delete(nodeKind);
    }
}
//# sourceMappingURL=readOperatorsAndOperands.js.map